"""Containers module."""

import contextlib
import copy as copy_module
import json
import sys
import importlib
import inspect
import warnings

try:
    import asyncio
except ImportError:
    asyncio = None

try:
    import yaml
except ImportError:
    yaml = None

import six

from . import providers, errors
from .providers cimport __is_future_or_coroutine


if sys.version_info[:2] >= (3, 6):
    from .wiring import wire, unwire
else:
    def wire(*args, **kwargs):
        raise NotImplementedError("Wiring requires Python 3.6 or above")

    def unwire(*args, **kwargs):
        raise NotImplementedError("Wiring requires Python 3.6 or above")

if sys.version_info[:2] == (3, 5):
    warnings.warn(
        "Dependency Injector will drop support of Python 3.5 after Jan 1st of 2022. "
        "This does not mean that there will be any immediate breaking changes, "
        "but tests will no longer be executed on Python 3.5, and bugs will not be addressed.",
        category=DeprecationWarning,
    )


class WiringConfiguration:
    """Container wiring configuration."""

    def __init__(self, modules=None, packages=None, from_package=None, auto_wire=True):
        self.modules = [*modules] if modules else []
        self.packages = [*packages] if packages else []
        self.from_package = from_package
        self.auto_wire = auto_wire

    def __deepcopy__(self, memo=None):
        return self.__class__(self.modules, self.packages, self.from_package, self.auto_wire)


class Container(object):
    """Abstract container."""


class DynamicContainer(Container):
    """Dynamic inversion of control container.

    .. code-block:: python

        services = DynamicContainer()
        services.auth = providers.Factory(AuthService)
        services.users = providers.Factory(UsersService,
                                           auth_service=services.auth)

    .. py:attribute:: providers

        Read-only dictionary of all providers.

        :type: dict[str, :py:class:`dependency_injector.providers.Provider`]

    .. py:attribute:: overridden

        Tuple of overriding containers.

        :type: tuple[:py:class:`DynamicContainer`]

    .. py:attribute:: provider_type

        Type of providers that could be placed in container.

        :type: type
    """

    __IS_CONTAINER__ = True

    def __init__(self):
        """Initializer.

        :rtype: None
        """
        self.provider_type = providers.Provider
        self.providers = {}
        self.overridden = tuple()
        self.parent = None
        self.declarative_parent = None
        self.wiring_config = WiringConfiguration()
        self.wired_to_modules = []
        self.wired_to_packages = []
        self.__self__ = providers.Self(self)
        super(DynamicContainer, self).__init__()

    def __deepcopy__(self, memo):
        """Create and return full copy of container."""
        copied = memo.get(id(self))
        if copied is not None:
            return copied

        copied = self.__class__()
        memo[id(self)] = copied

        copied.__self__ = providers.deepcopy(self.__self__, memo)
        for name in copied.__self__.alt_names:
            copied.set_provider(name, copied.__self__)

        copied.provider_type = providers.Provider
        copied.overridden = providers.deepcopy(self.overridden, memo)
        copied.wiring_config = copy_module.deepcopy(self.wiring_config, memo)
        copied.declarative_parent = self.declarative_parent

        for name, provider in providers.deepcopy(self.providers, memo).items():
            copied.set_provider(name, provider)

        copied.parent = providers.deepcopy(self.parent, memo)

        return copied

    def __setattr__(self, name, value):
        """Set instance attribute.

        If value of attribute is provider, it will be added into providers
        dictionary.

        :param name: Attribute name
        :type name: object

        :param value: Attribute value
        :type value: object

        :rtype: None
        """
        if isinstance(value, providers.Provider) \
                and not isinstance(value, providers.Self) \
                and name != "parent":
            _check_provider_type(self, value)

            self.providers[name] = value

            if isinstance(value, providers.CHILD_PROVIDERS):
                value.assign_parent(self)

        super(DynamicContainer, self).__setattr__(name, value)

    def __delattr__(self, name):
        """Delete instance attribute.

        If value of attribute is provider, it will be deleted from providers
        dictionary.

        :param name: Attribute name
        :type name: object

        :rtype: None
        """
        if name in self.providers:
            del self.providers[name]
        super(DynamicContainer, self).__delattr__(name)

    @property
    def dependencies(self):
        """Return dependency providers dictionary.

        Dependency providers can be both of :py:class:`dependency_injector.providers.Dependency` and
        :py:class:`dependency_injector.providers.DependenciesContainer`.

        :rtype:
            dict[str, :py:class:`dependency_injector.providers.Provider`]
        """
        return {
            name: provider
            for name, provider in self.providers.items()
            if isinstance(provider, (providers.Dependency, providers.DependenciesContainer))
        }

    def traverse(self, types=None):
        """Return providers traversal generator."""
        yield from providers.traverse(*self.providers.values(), types=types)

    def set_providers(self, **providers):
        """Set container providers.

        :param providers: Dictionary of providers
        :type providers:
            dict[object, :py:class:`dependency_injector.providers.Provider`]

        :rtype: None
        """
        for name, provider in six.iteritems(providers):
            setattr(self, name, provider)

    def set_provider(self, name, provider):
        """Set container provider.

        :param name: Provider name
        :type name: str

        :param provider: Provider
        :type provider: :py:class:`dependency_injector.providers.Provider`

        :rtype: None
        """
        setattr(self, name, provider)

    def override(self, object overriding):
        """Override current container by overriding container.

        :param overriding: Overriding container.
        :type overriding: :py:class:`DynamicContainer`

        :raise: :py:exc:`dependency_injector.errors.Error` if trying to
                override container by itself

        :rtype: None
        """
        if overriding is self:
            raise errors.Error("Container {0} could not be overridden "
                               "with itself".format(self))

        self.overridden += (overriding,)

        for name, provider in six.iteritems(overriding.providers):
            try:
                getattr(self, name).override(provider)
            except AttributeError:
                pass

    def override_providers(self, **overriding_providers):
        """Override container providers.

        :param overriding_providers: Dictionary of providers
        :type overriding_providers:
            dict[str, :py:class:`dependency_injector.providers.Provider`]

        :rtype: None
        """
        overridden_providers = []
        for name, overriding_provider in six.iteritems(overriding_providers):
            container_provider = getattr(self, name)
            container_provider.override(overriding_provider)
            overridden_providers.append(container_provider)
        return ProvidersOverridingContext(self, overridden_providers)

    def reset_last_overriding(self):
        """Reset last overriding provider for each container providers.

        :rtype: None
        """
        if not self.overridden:
            raise errors.Error("Container {0} is not overridden".format(self))

        self.overridden = self.overridden[:-1]

        for provider in six.itervalues(self.providers):
            provider.reset_last_overriding()

    def reset_override(self):
        """Reset all overridings for each container providers.

        :rtype: None
        """
        self.overridden = tuple()

        for provider in six.itervalues(self.providers):
            provider.reset_override()

    def is_auto_wiring_enabled(self):
        """Check if auto wiring is needed."""
        return self.wiring_config.auto_wire is True

    def wire(self, modules=None, packages=None, from_package=None):
        """Wire container providers with provided packages and modules.

        :rtype: None
        """
        if modules is None and self.wiring_config.modules:
            modules = self.wiring_config.modules
        if packages is None and self.wiring_config.packages:
            packages = self.wiring_config.packages

        modules = [*modules] if modules else []
        packages = [*packages] if packages else []

        if _any_relative_string_imports_in(modules) or _any_relative_string_imports_in(packages):
            if from_package is None:
                if self.wiring_config.from_package is not None:
                    from_package = self.wiring_config.from_package
                elif self.declarative_parent is not None \
                        and (self.wiring_config.modules or self.wiring_config.packages):
                    with contextlib.suppress(Exception):
                        from_package = _resolve_package_name_from_cls(self.declarative_parent)
                else:
                    with contextlib.suppress(Exception):
                        from_package = _resolve_calling_package_name()

        modules = _resolve_string_imports(modules, from_package)
        packages = _resolve_string_imports(packages, from_package)

        if not modules and not packages:
            return

        wire(
            container=self,
            modules=modules,
            packages=packages,
        )

        if modules:
            self.wired_to_modules.extend(modules)
        if packages:
            self.wired_to_packages.extend(packages)

    def unwire(self):
        """Unwire container providers from previously wired packages and modules."""
        unwire(
            modules=self.wired_to_modules,
            packages=self.wired_to_packages,
        )

        self.wired_to_modules.clear()
        self.wired_to_packages.clear()

    def init_resources(self):
        """Initialize all container resources."""
        futures = []

        for provider in self.traverse(types=[providers.Resource]):
            resource = provider.init()

            if __is_future_or_coroutine(resource):
                futures.append(resource)

        if futures:
            return asyncio.gather(*futures)

    def shutdown_resources(self):
        """Shutdown all container resources."""
        def _independent_resources(resources):
            for resource in resources:
                for other_resource in resources:
                    if not other_resource.initialized:
                        continue
                    if resource in other_resource.related:
                        break
                else:
                    yield resource

        async def _async_ordered_shutdown(resources):
            while any(resource.initialized for resource in resources):
                resources_to_shutdown = list(_independent_resources(resources))
                if not resources_to_shutdown:
                    raise RuntimeError("Unable to resolve resources shutdown order")
                futures = []
                for resource in resources_to_shutdown:
                    result = resource.shutdown()
                    if __is_future_or_coroutine(result):
                        futures.append(result)
                await asyncio.gather(*futures)

        def _sync_ordered_shutdown(resources):
            while any(resource.initialized for resource in resources):
                resources_to_shutdown = list(_independent_resources(resources))
                if not resources_to_shutdown:
                    raise RuntimeError("Unable to resolve resources shutdown order")
                for resource in resources_to_shutdown:
                    resource.shutdown()

        resources = list(self.traverse(types=[providers.Resource]))
        if any(resource.is_async_mode_enabled() for resource in resources):
            return _async_ordered_shutdown(resources)
        else:
            return _sync_ordered_shutdown(resources)

    def load_config(self):
        """Load configuration."""
        config: providers.Configuration
        for config in self.traverse(types=[providers.Configuration]):
            config.load()

    def apply_container_providers_overridings(self):
        """Apply container providers overridings."""
        for provider in self.traverse(types=[providers.Container]):
            provider.apply_overridings()

    def reset_singletons(self):
        """Reset container singletons."""
        for provider in self.traverse(types=[providers.BaseSingleton]):
            provider.reset()
        return SingletonResetContext(self)

    def check_dependencies(self):
        """Check if container dependencies are defined.

        If any dependency is undefined, raises an error.
        """
        undefined = [
            dependency
            for dependency in self.traverse(types=[providers.Dependency])
            if not dependency.is_defined
        ]

        if not undefined:
            return

        container_name = self.parent_name if self.parent_name else self.__class__.__name__
        undefined_names = [
            f"\"{dependency.parent_name if dependency.parent_name else dependency}\""
            for dependency in undefined
        ]
        raise errors.Error(
            f"Container \"{container_name}\" has undefined dependencies: "
            f"{', '.join(undefined_names)}",
        )

    def from_schema(self, schema):
        """Build container providers from schema."""
        from .schema import build_schema
        for name, provider in build_schema(schema).items():
            self.set_provider(name, provider)

    def from_yaml_schema(self, filepath, loader=None):
        """Build container providers from YAML schema.

        You can specify type of loader as a second argument. By default, method
        uses ``SafeLoader``.
        """
        if yaml is None:
            raise errors.Error(
                "Unable to load yaml schema - PyYAML is not installed. "
                "Install PyYAML or install Dependency Injector with yaml extras: "
                "\"pip install dependency-injector[yaml]\""
            )

        if loader is None:
            loader = yaml.SafeLoader

        with open(filepath) as file:
            schema = yaml.load(file, loader)

        self.from_schema(schema)

    def from_json_schema(self, filepath):
        """Build container providers from JSON schema."""
        with open(filepath) as file:
            schema = json.load(file)
        self.from_schema(schema)

    def resolve_provider_name(self, provider):
        """Try to resolve provider name."""
        for provider_name, container_provider in self.providers.items():
            if container_provider is provider:
                return provider_name
        else:
            raise errors.Error(f"Can not resolve name for provider \"{provider}\"")

    @property
    def parent_name(self):
        """Return parent name."""
        if self.parent:
            return self.parent.parent_name

        if self.declarative_parent:
            return self.declarative_parent.__name__

        return None

    def assign_parent(self, parent):
        """Assign parent."""
        self.parent = parent


class DeclarativeContainerMetaClass(type):
    """Declarative inversion of control container meta class."""

    def __new__(type mcs, str class_name, tuple bases, dict attributes):
        """Declarative container class factory."""
        self = mcs.__fetch_self(attributes)
        if self is None:
            self = providers.Self()

        containers = {
            name: container
            for name, container in six.iteritems(attributes)
            if is_container(container)
        }

        cls_providers = {
            name: provider
            for name, provider in six.iteritems(attributes)
            if isinstance(provider, providers.Provider) and not isinstance(provider, providers.Self)
        }

        inherited_providers = {
            name: provider
            for base in bases
            if is_container(base) and base is not DynamicContainer
            for name, provider in six.iteritems(base.providers)
        }

        all_providers = {}
        all_providers.update(inherited_providers)
        all_providers.update(cls_providers)

        wiring_config = attributes.get("wiring_config")
        if wiring_config is None:
            wiring_config = WiringConfiguration()
        if wiring_config is not None and not isinstance(wiring_config, WiringConfiguration):
            raise errors.Error(
                "Wiring configuration should be an instance of WiringConfiguration, "
                "instead got {0}".format(wiring_config)
            )

        attributes["containers"] = containers
        attributes["inherited_providers"] = inherited_providers
        attributes["cls_providers"] = cls_providers
        attributes["providers"] = all_providers
        attributes["wiring_config"] = wiring_config

        cls = <type>type.__new__(mcs, class_name, bases, attributes)

        self.set_container(cls)
        cls.__self__ = self

        for provider in six.itervalues(cls.providers):
            _check_provider_type(cls, provider)

        for provider in six.itervalues(cls.cls_providers):
            if isinstance(provider, providers.CHILD_PROVIDERS):
                provider.assign_parent(cls)

        return cls

    def __setattr__(cls, name, value):
        """Set class attribute.

        If value of attribute is provider, it will be added into providers
        dictionary.

        :param name: Attribute name
        :type name: object

        :param value: Attribute value
        :type value: object

        :rtype: None
        """
        if isinstance(value, providers.Provider) and name != "__self__":
            _check_provider_type(cls, value)

            if isinstance(value, providers.CHILD_PROVIDERS):
                value.assign_parent(cls)

            cls.providers[name] = value
            cls.cls_providers[name] = value
        super(DeclarativeContainerMetaClass, cls).__setattr__(name, value)

    def __delattr__(cls, name):
        """Delete class attribute.

        If value of attribute is provider, it will be deleted from providers
        dictionary.

        :param name: Attribute name
        :type name: object

        :rtype: None
        """
        if name in cls.providers and name in cls.cls_providers:
            del cls.providers[name]
            del cls.cls_providers[name]
        super(DeclarativeContainerMetaClass, cls).__delattr__(name)

    @property
    def dependencies(cls):
        """Return dependency providers dictionary.

        Dependency providers can be both of :py:class:`dependency_injector.providers.Dependency` and
        :py:class:`dependency_injector.providers.DependenciesContainer`.

        :rtype:
            dict[str, :py:class:`dependency_injector.providers.Provider`]
        """
        return {
            name: provider
            for name, provider in cls.providers.items()
            if isinstance(provider, (providers.Dependency, providers.DependenciesContainer))
        }

    def traverse(cls, types=None):
        """Return providers traversal generator."""
        yield from providers.traverse(*cls.providers.values(), types=types)

    def resolve_provider_name(cls, provider):
        """Try to resolve provider name."""
        for provider_name, container_provider in cls.providers.items():
            if container_provider is provider:
                return provider_name
        else:
            raise errors.Error(f"Can not resolve name for provider \"{provider}\"")

    @property
    def parent_name(cls):
        """Return parent name."""
        return cls.__name__

    @staticmethod
    def __fetch_self(attributes):
        self = None
        alt_names = []

        for name, value in attributes.items():
            if not isinstance(value, providers.Self):
                continue

            if self is not None and value is not self:
                raise errors.Error("Container can have only one \"Self\" provider")

            if name != "__self__":
                alt_names.append(name)

            self = value

        if self:
            self.set_alt_names(alt_names)

        return self


@six.add_metaclass(DeclarativeContainerMetaClass)
class DeclarativeContainer(Container):
    """Declarative inversion of control container.

    .. code-block:: python

        class Services(DeclarativeContainer):
            auth = providers.Factory(AuthService)
            users = providers.Factory(UsersService,
                                      auth_service=auth)
    """

    __IS_CONTAINER__ = True

    provider_type = providers.Provider
    """Type of providers that could be placed in container.

    :type: type
    """

    instance_type = DynamicContainer
    """Type of container that is returned on instantiating declarative
    container.

    :type: type
    """

    containers = dict()
    """Read-only dictionary of all nested containers.

    :type: dict[str, :py:class:`DeclarativeContainer`]
    """

    providers = dict()
    """Read-only dictionary of all providers.

    :type: dict[str, :py:class:`dependency_injector.providers.Provider`]
    """

    wiring_config = WiringConfiguration()
    """Wiring configuration.

    :type: WiringConfiguration
    """

    auto_load_config = True
    """Automatically load configuration when the container is created.

    :type: bool
    """

    cls_providers = dict()
    """Read-only dictionary of current container providers.

    :type: dict[str, :py:class:`dependency_injector.providers.Provider`]
    """

    inherited_providers = dict()
    """Read-only dictionary of inherited providers.

    :type: dict[str, :py:class:`dependency_injector.providers.Provider`]
    """

    overridden = tuple()
    """Tuple of overriding containers.

    :type: tuple[:py:class:`DeclarativeContainer`]
    """

    __self__ = None
    """Provider that provides current container.

    :type: :py:class:`dependency_injector.providers.Provider`
    """

    def __new__(cls, **overriding_providers):
        """Constructor.

        :return: Dynamic container with copy of all providers.
        :rtype: :py:class:`DynamicContainer`
        """
        container = cls.instance_type()
        container.provider_type = cls.provider_type
        container.wiring_config = copy_module.deepcopy(cls.wiring_config)
        container.declarative_parent = cls

        copied_providers = providers.deepcopy({ **cls.providers, **{"@@self@@": cls.__self__}})
        copied_self = copied_providers.pop("@@self@@")
        copied_self.set_container(container)

        container.__self__ = copied_self
        for name in copied_self.alt_names:
            container.set_provider(name, copied_self)

        for name, provider in copied_providers.items():
            container.set_provider(name, provider)

        if cls.auto_load_config:
            container.load_config()

        container.override_providers(**overriding_providers)
        container.apply_container_providers_overridings()

        if container.is_auto_wiring_enabled():
            container.wire()

        return container

    @classmethod
    def override(cls, object overriding):
        """Override current container by overriding container.

        :param overriding: Overriding container.
        :type overriding: :py:class:`DeclarativeContainer`

        :raise: :py:exc:`dependency_injector.errors.Error` if trying to
                override container by itself or its subclasses

        :rtype: None
        """
        if issubclass(cls, overriding):
            raise errors.Error("Container {0} could not be overridden "
                               "with itself or its subclasses".format(cls))

        cls.overridden += (overriding,)

        for name, provider in six.iteritems(overriding.cls_providers):
            try:
                getattr(cls, name).override(provider)
            except AttributeError:
                pass

    @classmethod
    def reset_last_overriding(cls):
        """Reset last overriding provider for each container providers.

        :rtype: None
        """
        if not cls.overridden:
            raise errors.Error("Container {0} is not overridden".format(cls))

        cls.overridden = cls.overridden[:-1]

        for provider in six.itervalues(cls.providers):
            provider.reset_last_overriding()

    @classmethod
    def reset_override(cls):
        """Reset all overridings for each container providers.

        :rtype: None
        """
        cls.overridden = tuple()

        for provider in six.itervalues(cls.providers):
            provider.reset_override()


class SingletonResetContext:

    def __init__(self, container):
        self._container = container

    def __enter__(self):
        return self._container

    def __exit__(self, *_):
        self._container.reset_singletons()



class ProvidersOverridingContext:

    def __init__(self, container, overridden_providers):
        self._container = container
        self._overridden_providers = overridden_providers

    def __enter__(self):
        return self._container

    def __exit__(self, *_):
        for provider in self._overridden_providers:
            provider.reset_last_overriding()


def override(object container):
    """:py:class:`DeclarativeContainer` overriding decorator.

    :param container: Container that should be overridden by decorated
                      container.
    :type container: :py:class:`DeclarativeContainer`

    :return: Declarative container overriding decorator.
    :rtype: callable(:py:class:`DeclarativeContainer`)
    """
    def _decorator(object overriding_container):
        """Overriding decorator."""
        container.override(overriding_container)
        return overriding_container
    return _decorator


def copy(object base_container):
    """:py:class:`DeclarativeContainer` copying decorator.

    This decorator copies all providers from provided container to decorated one.
    If one of the decorated container providers matches to source container
    providers by name, it would be replaced by reference.

    :param base_container: Container that should be copied by decorated container.
    :type base_container: :py:class:`DeclarativeContainer`

    :return: Declarative container copying decorator.
    :rtype: callable(:py:class:`DeclarativeContainer`)
    """
    def _get_memo_for_matching_names(new_providers, base_providers):
        memo = {}
        for new_provider_name, new_provider in six.iteritems(new_providers):
            if new_provider_name not in base_providers:
                continue
            source_provider = base_providers[new_provider_name]
            memo[id(source_provider)] = new_provider

            if hasattr(new_provider, "providers") and hasattr(source_provider, "providers"):
                sub_memo = _get_memo_for_matching_names(new_provider.providers, source_provider.providers)
                memo.update(sub_memo)
        return memo

    def _decorator(new_container):
        memo = {}
        memo.update(_get_memo_for_matching_names(new_container.cls_providers, base_container.providers))

        new_providers = {}
        new_providers.update(providers.deepcopy(base_container.providers, memo))
        new_providers.update(providers.deepcopy(new_container.cls_providers, memo))

        for name, provider in six.iteritems(new_providers):
            setattr(new_container, name, provider)
        return new_container

    return _decorator


cpdef bint is_container(object instance):
    """Check if instance is container instance.

    :param instance: Instance to be checked.
    :type instance: object

    :rtype: bool
    """
    return getattr(instance, "__IS_CONTAINER__", False) is True


cpdef object _check_provider_type(object container, object provider):
    if not isinstance(provider, container.provider_type):
        raise errors.Error("{0} can contain only {1} "
                           "instances".format(container, container.provider_type))


cpdef bint _any_relative_string_imports_in(object modules):
    for module in modules:
        if not isinstance(module, str):
            continue
        if module.startswith("."):
            return True
    else:
        return False


cpdef list _resolve_string_imports(object modules, object from_package):
    return [
        importlib.import_module(module, from_package) if isinstance(module, str) else module
        for module in modules
    ]


cpdef object _resolve_calling_package_name():
    stack = inspect.stack()
    pre_last_frame = stack[0]
    module = inspect.getmodule(pre_last_frame[0])
    return module.__package__


cpdef object _resolve_package_name_from_cls(cls):
    module = importlib.import_module(cls.__module__)
    return module.__package__



from typing import Callable, Optional, Tuple, Any, Dict, Type

from dependency_injector import providers


class Animal:
    ...


class Cat(Animal):

    def __init__(self, *_, **__): ...

    @classmethod
    def create(cls) -> Animal:
        return cls()


# Test 1: to check the return type (class)
provider1 = providers.Factory(Cat)
animal1: Animal = provider1(1, 2, 3, b="1", c=2, e=0.0)

# Test 2: to check the return type (class factory method)
provider2 = providers.Factory(Cat.create)
animal2: Animal = provider2()

# Test 3: to check the .override() method
provider3 = providers.Factory(Animal)
with provider3.override(providers.Factory(Cat)):
    provider3()

# Test 4: to check the .args, .kwargs, .attributes attributes
provider4 = providers.Factory(Animal)
args4: Tuple[Any] = provider4.args
kwargs4: Dict[str, Any] = provider4.kwargs
attributes4: Dict[str, Any] = provider4.attributes

# Test 5: to check the provided instance interface
provider5 = providers.Factory(Animal)
provided5: providers.ProvidedInstance = provider5.provided
attr_getter5: providers.AttributeGetter = provider5.provided.attr
item_getter5: providers.ItemGetter = provider5.provided["item"]
method_caller5: providers.MethodCaller = provider5.provided.method.call(123, arg=324)

# Test 6: to check the DelegatedFactory
provider6 = providers.DelegatedFactory(Cat)
animal6: Animal = provider6(1, 2, 3, b="1", c=2, e=0.0)

# Test 7: to check the AbstractFactory
provider7 = providers.AbstractFactory(Animal)
provider7.override(providers.Factory(Cat))
animal7: Animal = provider7(1, 2, 3, b="1", c=2, e=0.0)

# Test 8: to check the FactoryDelegate __init__
provider8 = providers.FactoryDelegate(providers.Factory(object))

# Test 9: to check FactoryAggregate provider
provider9: providers.FactoryAggregate[str] = providers.FactoryAggregate(
    a=providers.Factory(str, "str1"),
    b=providers.Factory(str, "str2"),
)
factory_a_9: providers.Factory[str] = provider9.a
factory_b_9: providers.Factory[str] = provider9.b
val9: str = provider9("a")

provider9_set_non_string_keys: providers.FactoryAggregate[str] = providers.FactoryAggregate()
provider9_set_non_string_keys.set_factories({Cat: providers.Factory(str, "str")})
factory_set_non_string_9: providers.Factory[str] = provider9_set_non_string_keys.factories[Cat]

provider9_new_non_string_keys: providers.FactoryAggregate[str] = providers.FactoryAggregate(
    {Cat: providers.Factory(str, "str")},
)
factory_new_non_string_9: providers.Factory[str] = provider9_new_non_string_keys.factories[Cat]

provider9_no_explicit_typing = providers.FactoryAggregate(a=providers.Factory(str, "str"))
provider9_no_explicit_typing_factory: providers.Factory[str] = provider9_no_explicit_typing.factories["a"]
provider9_no_explicit_typing_object: str = provider9_no_explicit_typing("a")

# Test 10: to check the explicit typing
factory10: providers.Provider[Animal] = providers.Factory(Cat)
animal10: Animal = factory10()

# Test 11: to check the return type with await
provider11 = providers.Factory(Cat)
async def _async11() -> None:
    animal1: Animal = await provider11(1, 2, 3, b="1", c=2, e=0.0)  # type: ignore
    animal2: Animal = await provider11.async_(1, 2, 3, b="1", c=2, e=0.0)

# Test 12: to check class type from .provides
provider12 = providers.Factory(Cat)
provided_cls12: Type[Animal] = provider12.cls
assert issubclass(provided_cls12, Animal)
provided_provides12: Optional[Callable[..., Animal]] = provider12.provides
assert provided_provides12 is not None and provided_provides12() == Cat()

# Test 13: to check class from .provides with explicit typevar
provider13 = providers.Factory[Animal](Cat)
provided_cls13: Type[Animal] = provider13.cls
assert issubclass(provided_cls13, Animal)
provided_provides13: Optional[Callable[..., Animal]] = provider13.provides
assert provided_provides13 is not None and provided_provides13() == Cat()

# Test 14: to check string imports
provider14: providers.Factory[dict] = providers.Factory("builtins.dict")
provider14.set_provides("builtins.dict")



"""Tests for container with custom string classes as attribute names.

See: https://github.com/ets-labs/python-dependency-injector/issues/479
"""

from dependency_injector import containers, providers
from pytest import fixture, raises


class CustomString(str):
    pass


class CustomClass:
    thing = None


@fixture
def container():
    return containers.DynamicContainer()


@fixture
def provider():
    return providers.Provider()


def test_setattr(container, provider):
    setattr(container, CustomString("test_attr"), provider)
    assert container.test_attr is provider


def test_delattr(container, provider):
    setattr(container, CustomString("test_attr"), provider)
    delattr(container, CustomString("test_attr"))
    with raises(AttributeError):
        container.test_attr


def test_set_provider(container, provider):
    container.set_provider(CustomString("test_attr"), provider)
    assert container.test_attr is provider


def test_set_providers(container, provider):
    container.set_providers(**{CustomString("test_attr"): provider})
    assert container.test_attr is provider



"""Providers module."""

try:
    import asyncio
except ImportError:
    asyncio = None

import functools

cimport cython


cdef int ASYNC_MODE_UNDEFINED
cdef int ASYNC_MODE_ENABLED
cdef int ASYNC_MODE_DISABLED

cdef set __iscoroutine_typecache
cdef tuple __COROUTINE_TYPES


# Base providers
cdef class Provider(object):
    cdef tuple _overridden
    cdef Provider _last_overriding
    cdef tuple _overrides
    cdef int _async_mode

    cpdef bint is_async_mode_enabled(self)
    cpdef bint is_async_mode_disabled(self)
    cpdef bint is_async_mode_undefined(self)

    cpdef object _provide(self, tuple args, dict kwargs)
    cpdef void _copy_overridings(self, Provider copied, dict memo)


cdef class Object(Provider):
    cdef object _provides

    cpdef object _provide(self, tuple args, dict kwargs)


cdef class Self(Provider):
    cdef object _container
    cdef tuple _alt_names


cdef class Delegate(Provider):
    cdef object _provides

    cpdef object _provide(self, tuple args, dict kwargs)


cdef class Aggregate(Provider):
    cdef dict _providers

    cdef Provider __get_provider(self, object provider_name)


cdef class Dependency(Provider):
    cdef object _instance_of
    cdef object _default
    cdef object _parent


cdef class ExternalDependency(Dependency):
    pass


cdef class DependenciesContainer(Object):
    cdef dict _providers
    cdef object _parent

    cpdef object _override_providers(self, object container)


# Callable providers
cdef class Callable(Provider):
    cdef object _provides

    cdef tuple _args
    cdef int _args_len

    cdef tuple _kwargs
    cdef int _kwargs_len

    cpdef object _provide(self, tuple args, dict kwargs)


cdef class DelegatedCallable(Callable):
    pass


cdef class AbstractCallable(Callable):
    cpdef object _provide(self, tuple args, dict kwargs)


cdef class CallableDelegate(Delegate):
    pass


# Coroutine providers
cdef class Coroutine(Callable):
    pass


cdef class DelegatedCoroutine(Coroutine):
    pass


cdef class AbstractCoroutine(Coroutine):
    cpdef object _provide(self, tuple args, dict kwargs)


cdef class CoroutineDelegate(Delegate):
    pass


# Configuration providers
cdef class ConfigurationOption(Provider):
    cdef tuple _name
    cdef Configuration _root
    cdef dict _children
    cdef bint _required
    cdef object _cache


cdef class TypedConfigurationOption(Callable):
    pass


cdef class Configuration(Object):
    cdef str _name
    cdef bint __strict
    cdef dict _children
    cdef list _ini_files
    cdef list _yaml_files
    cdef list _json_files
    cdef list _pydantic_settings
    cdef object __weakref__


# Factory providers
cdef class Factory(Provider):
    cdef Callable _instantiator

    cdef tuple _attributes
    cdef int _attributes_len

    cpdef object _provide(self, tuple args, dict kwargs)


cdef class DelegatedFactory(Factory):
    pass


cdef class AbstractFactory(Factory):
    cpdef object _provide(self, tuple args, dict kwargs)


cdef class FactoryDelegate(Delegate):
    pass


cdef class FactoryAggregate(Aggregate):
    pass


# Singleton providers
cdef class BaseSingleton(Provider):
    cdef Factory _instantiator
    cdef object _storage


cdef class Singleton(BaseSingleton):

    cpdef object _provide(self, tuple args, dict kwargs)


cdef class DelegatedSingleton(Singleton):
    pass


cdef class ThreadSafeSingleton(BaseSingleton):
    cdef object _storage_lock

    cpdef object _provide(self, tuple args, dict kwargs)


cdef class DelegatedThreadSafeSingleton(ThreadSafeSingleton):
    pass


cdef class ThreadLocalSingleton(BaseSingleton):

    cpdef object _provide(self, tuple args, dict kwargs)


cdef class ContextLocalSingleton(BaseSingleton):

    cpdef object _provide(self, tuple args, dict kwargs)


cdef class DelegatedThreadLocalSingleton(ThreadLocalSingleton):
    pass


cdef class AbstractSingleton(BaseSingleton):
    pass


cdef class SingletonDelegate(Delegate):
    pass


# Miscellaneous providers

cdef class List(Provider):
    cdef tuple _args
    cdef int _args_len

    cpdef object _provide(self, tuple args, dict kwargs)


cdef class Dict(Provider):
    cdef tuple _kwargs
    cdef int _kwargs_len

    cpdef object _provide(self, tuple args, dict kwargs)


cdef class Resource(Provider):
    cdef object _provides
    cdef bint _initialized
    cdef object _shutdowner
    cdef object _resource

    cdef tuple _args
    cdef int _args_len

    cdef tuple _kwargs
    cdef int _kwargs_len

    cpdef object _provide(self, tuple args, dict kwargs)


cdef class Container(Provider):
    cdef object _container_cls
    cdef dict _overriding_providers
    cdef object _container
    cdef object _parent

    cpdef object _provide(self, tuple args, dict kwargs)


cdef class Selector(Provider):
    cdef object _selector
    cdef dict _providers

    cpdef object _provide(self, tuple args, dict kwargs)

# Provided instance

cdef class ProvidedInstance(Provider):
    cdef object _provides

    cpdef object _provide(self, tuple args, dict kwargs)


cdef class AttributeGetter(Provider):
    cdef object _provides
    cdef object _name

    cpdef object _provide(self, tuple args, dict kwargs)


cdef class ItemGetter(Provider):
    cdef object _provides
    cdef object _name

    cpdef object _provide(self, tuple args, dict kwargs)


cdef class MethodCaller(Provider):
    cdef object _provides
    cdef tuple _args
    cdef int _args_len
    cdef tuple _kwargs
    cdef int _kwargs_len

    cpdef object _provide(self, tuple args, dict kwargs)


# Injections
cdef class Injection(object):
    cdef object _value
    cdef int _is_provider
    cdef int _is_delegated
    cdef int _call


cdef class PositionalInjection(Injection):
    pass


cdef class NamedInjection(Injection):
    cdef object _name


cpdef tuple parse_positional_injections(tuple args)


cpdef tuple parse_named_injections(dict kwargs)


# Utils
cdef class OverridingContext(object):
    cdef Provider _overridden
    cdef Provider _overriding


cdef class BaseSingletonResetContext(object):
    cdef object _singleton


cdef class SingletonResetContext(BaseSingletonResetContext):
    pass


cdef class SingletonFullResetContext(BaseSingletonResetContext):
    pass


cdef object CLASS_TYPES


cpdef bint is_provider(object instance)


cpdef object ensure_is_provider(object instance)


cpdef bint is_delegated(object instance)


cpdef str represent_provider(object provider, object provides)


cpdef bint is_container_instance(object instance)


cpdef bint is_container_class(object instance)


cpdef object deepcopy(object instance, dict memo=*)


# Inline helper functions
cdef inline object __get_name(NamedInjection self):
    return self._name


cdef inline object __get_value(Injection self):
    if self._call == 0:
        return self._value
    return self._value()


cdef inline object __get_value_kwargs(Injection self, dict kwargs):
    if self._call == 0:
        return self._value
    return self._value(**kwargs)


cdef inline tuple __separate_prefixed_kwargs(dict kwargs):
    cdef dict plain_kwargs = {}
    cdef dict prefixed_kwargs = {}

    for key, value in kwargs.items():
        if "__" not in key:
            plain_kwargs[key] = value
            continue

        index = key.index("__")
        prefix, name = key[:index], key[index+2:]

        if prefix not in prefixed_kwargs:
            prefixed_kwargs[prefix] = {}
        prefixed_kwargs[prefix][name] = value

    return plain_kwargs, prefixed_kwargs


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline object __provide_positional_args(
        tuple args,
        tuple inj_args,
        int inj_args_len,
        int async_mode,
):
    cdef int index
    cdef list positional_args = []
    cdef list future_args = []
    cdef PositionalInjection injection
    cdef object value

    if inj_args_len == 0:
        return args

    for index in range(inj_args_len):
        injection = <PositionalInjection>inj_args[index]
        value = __get_value(injection)
        positional_args.append(value)

        if async_mode != ASYNC_MODE_DISABLED and __is_future_or_coroutine(value):
            future_args.append((index, value))

    positional_args.extend(args)

    if future_args:
        return __combine_future_injections(positional_args, future_args)

    return positional_args


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline object __provide_keyword_args(
        dict kwargs,
        tuple inj_kwargs,
        int inj_kwargs_len,
        int async_mode,
):
    cdef int index
    cdef object name
    cdef object value
    cdef dict prefixed = {}
    cdef list future_kwargs = []
    cdef NamedInjection kw_injection

    if len(kwargs) == 0:
        for index in range(inj_kwargs_len):
            kw_injection = <NamedInjection>inj_kwargs[index]
            name = __get_name(kw_injection)
            value = __get_value(kw_injection)
            kwargs[name] = value
            if async_mode != ASYNC_MODE_DISABLED and __is_future_or_coroutine(value):
                future_kwargs.append((name, value))
    else:
        kwargs, prefixed = __separate_prefixed_kwargs(kwargs)


        for index in range(inj_kwargs_len):
            kw_injection = <NamedInjection>inj_kwargs[index]
            name = __get_name(kw_injection)

            if name in kwargs:
                continue

            if name in prefixed:
                value = __get_value_kwargs(kw_injection, prefixed[name])
            else:
                value = __get_value(kw_injection)

            kwargs[name] = value
            if async_mode != ASYNC_MODE_DISABLED and __is_future_or_coroutine(value):
                future_kwargs.append((name, value))

    if future_kwargs:
        return __combine_future_injections(kwargs, future_kwargs)

    return kwargs


cdef inline object __combine_future_injections(object injections, list future_injections):
    future_result = asyncio.Future()

    injections_ready = asyncio.gather(*[value for _, value in future_injections])
    injections_ready.add_done_callback(
        functools.partial(
            __async_prepare_args_kwargs_callback,
            future_result,
            injections,
            future_injections,
        ),
    )
    asyncio.ensure_future(injections_ready)

    return future_result


cdef inline void __async_prepare_args_kwargs_callback(
        object future_result,
        object args,
        object future_args_kwargs,
        object future,
):
    try:
        result = future.result()
        for value, (key, _) in zip(result, future_args_kwargs):
            args[key] = value
    except Exception as exception:
        future_result.set_exception(exception)
    else:
        future_result.set_result(args)


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline object __provide_attributes(tuple attributes, int attributes_len):
    cdef NamedInjection attr_injection
    cdef dict attribute_injections = {}
    cdef list future_attributes = []

    for index in range(attributes_len):
        attr_injection = <NamedInjection>attributes[index]
        name = __get_name(attr_injection)
        value = __get_value(attr_injection)
        attribute_injections[name] = value
        if __is_future_or_coroutine(value):
            future_attributes.append((name, value))

    if future_attributes:
        return __combine_future_injections(attribute_injections, future_attributes)

    return attribute_injections


cdef inline object __async_inject_attributes(future_instance, future_attributes):
    future_result = asyncio.Future()

    attributes_ready = asyncio.gather(future_instance, future_attributes)
    attributes_ready.add_done_callback(
        functools.partial(
            __async_inject_attributes_callback,
            future_result,
        ),
    )
    asyncio.ensure_future(attributes_ready)

    return future_result


cdef inline void __async_inject_attributes_callback(object future_result, object future):
    try:
        instance, attributes = future.result()

        for name, value in attributes.items():
            setattr(instance, name, value)
    except Exception as exception:
        future_result.set_exception(exception)
    else:
        future_result.set_result(instance)


cdef inline void __inject_attributes(object instance, dict attributes):
    for name, value in attributes.items():
        setattr(instance, name, value)


cdef inline object __call(
        object call,
        tuple context_args,
        tuple injection_args,
        int injection_args_len,
        dict context_kwargs,
        tuple injection_kwargs,
        int injection_kwargs_len,
        int async_mode,
):
    cdef object args = __provide_positional_args(
        context_args,
        injection_args,
        injection_args_len,
        async_mode,
    )
    cdef object kwargs = __provide_keyword_args(
        context_kwargs,
        injection_kwargs,
        injection_kwargs_len,
        async_mode,
    )

    if async_mode == ASYNC_MODE_DISABLED:
        return call(*args, **kwargs)

    cdef bint is_future_args = __is_future_or_coroutine(args)
    cdef bint is_future_kwargs = __is_future_or_coroutine(kwargs)

    if is_future_args or is_future_kwargs:
        future_args = args if is_future_args else __future_result(args)
        future_kwargs = kwargs if is_future_kwargs else __future_result(kwargs)

        future_result = asyncio.Future()

        args_kwargs_ready = asyncio.gather(future_args, future_kwargs)
        args_kwargs_ready.add_done_callback(
            functools.partial(
                __async_call_callback,
                future_result,
                call,
            ),
        )
        asyncio.ensure_future(args_kwargs_ready)

        return future_result

    return call(*args, **kwargs)


cdef inline void __async_call_callback(object future_result, object call, object future):
    try:
        args, kwargs = future.result()
        result = call(*args, **kwargs)
    except Exception as exception:
        future_result.set_exception(exception)
    else:
        if __is_future_or_coroutine(result):
            result = asyncio.ensure_future(result)
            result.add_done_callback(functools.partial(__async_result_callback, future_result))
            return
        future_result.set_result(result)


cdef inline object __async_result_callback(object future_result, object future):
    try:
        result = future.result()
    except Exception as exception:
        future_result.set_exception(exception)
    else:
        future_result.set_result(result)


cdef inline object __callable_call(Callable self, tuple args, dict kwargs, ):
    return __call(
        self._provides,
        args,
        self._args,
        self._args_len,
        kwargs,
        self._kwargs,
        self._kwargs_len,
        self._async_mode,
    )


cdef inline object __factory_call(Factory self, tuple args, dict kwargs):
    cdef object instance

    instance = __call(
        self._instantiator._provides,
        args,
        self._instantiator._args,
        self._instantiator._args_len,
        kwargs,
        self._instantiator._kwargs,
        self._instantiator._kwargs_len,
        self._async_mode,
    )

    if self._attributes_len > 0:
        attributes = __provide_attributes(self._attributes, self._attributes_len)

        is_future_instance = __is_future_or_coroutine(instance)
        is_future_attributes = __is_future_or_coroutine(attributes)

        if is_future_instance or is_future_attributes:
            future_instance = instance if is_future_instance else __future_result(instance)
            future_attributes = attributes if is_future_attributes else __future_result(attributes)
            return __async_inject_attributes(future_instance, future_attributes)

        __inject_attributes(instance, attributes)

    return instance


cdef inline bint __is_future_or_coroutine(object instance):
    return __isfuture(instance) or __iscoroutine(instance)


cdef inline bint __isfuture(object obj):
    return hasattr(obj.__class__, "_asyncio_future_blocking") and obj._asyncio_future_blocking is not None


cdef inline bint __iscoroutine(object obj):
    if type(obj) in __iscoroutine_typecache:
        return True

    if isinstance(obj, __COROUTINE_TYPES):
        # Just in case we don't want to cache more than 100
        # positive types.  That shouldn't ever happen, unless
        # someone stressing the system on purpose.
        if len(__iscoroutine_typecache) < 100:
            __iscoroutine_typecache.add(type(obj))
        return True
    else:
        return False


cdef inline object __future_result(object instance):
    future_result = asyncio.Future()
    future_result.set_result(instance)
    return future_result



"""Dict provider tests."""

import sys

from dependency_injector import providers


def test_is_provider():
    assert providers.is_provider(providers.Dict()) is True


def test_provided_instance_provider():
    provider = providers.Dict()
    assert isinstance(provider.provided, providers.ProvidedInstance)


def test_init_with_non_string_keys():
    a1 = object()
    a2 = object()
    provider = providers.Dict({a1: "i1", a2: "i2"})

    dict1 = provider()
    dict2 = provider()

    assert dict1 == {a1: "i1", a2: "i2"}
    assert dict2 == {a1: "i1", a2: "i2"}

    assert dict1 is not dict2


def test_init_with_string_and_non_string_keys():
    a1 = object()
    provider = providers.Dict({a1: "i1"}, a2="i2")

    dict1 = provider()
    dict2 = provider()

    assert dict1 == {a1: "i1", "a2": "i2"}
    assert dict2 == {a1: "i1", "a2": "i2"}

    assert dict1 is not dict2


def test_call_with_init_keyword_args():
    provider = providers.Dict(a1="i1", a2="i2")

    dict1 = provider()
    dict2 = provider()

    assert dict1 == {"a1": "i1", "a2": "i2"}
    assert dict2 == {"a1": "i1", "a2": "i2"}

    assert dict1 is not dict2


def test_call_with_context_keyword_args():
    provider = providers.Dict(a1="i1", a2="i2")
    assert provider(a3="i3", a4="i4") == {"a1": "i1", "a2": "i2", "a3": "i3", "a4": "i4"}


def test_call_with_provider():
    provider = providers.Dict(
        a1=providers.Factory(str, "i1"),
        a2=providers.Factory(str, "i2"),
    )
    assert provider() == {"a1": "i1", "a2": "i2"}


def test_fluent_interface():
    provider = providers.Dict() \
        .add_kwargs(a1="i1", a2="i2")
    assert provider() == {"a1": "i1", "a2": "i2"}


def test_add_kwargs():
    provider = providers.Dict() \
        .add_kwargs(a1="i1") \
        .add_kwargs(a2="i2")
    assert provider.kwargs == {"a1": "i1", "a2": "i2"}


def test_add_kwargs_non_string_keys():
    a1 = object()
    a2 = object()
    provider = providers.Dict() \
        .add_kwargs({a1: "i1"}) \
        .add_kwargs({a2: "i2"})
    assert provider.kwargs == {a1: "i1", a2: "i2"}


def test_add_kwargs_string_and_non_string_keys():
    a2 = object()
    provider = providers.Dict() \
        .add_kwargs(a1="i1") \
        .add_kwargs({a2: "i2"})
    assert provider.kwargs == {"a1": "i1", a2: "i2"}


def test_set_kwargs():
    provider = providers.Dict() \
        .add_kwargs(a1="i1", a2="i2") \
        .set_kwargs(a3="i3", a4="i4")
    assert provider.kwargs == {"a3": "i3", "a4": "i4"}


def test_set_kwargs_non_string_keys():
    a3 = object()
    a4 = object()
    provider = providers.Dict() \
        .add_kwargs(a1="i1", a2="i2") \
        .set_kwargs({a3: "i3", a4: "i4"})
    assert provider.kwargs == {a3: "i3", a4: "i4"}


def test_set_kwargs_string_and_non_string_keys():
    a3 = object()
    provider = providers.Dict() \
        .add_kwargs(a1="i1", a2="i2") \
        .set_kwargs({a3: "i3"}, a4="i4")
    assert provider.kwargs == {a3: "i3", "a4": "i4"}


def test_clear_kwargs():
    provider = providers.Dict() \
        .add_kwargs(a1="i1", a2="i2") \
        .clear_kwargs()
    assert provider.kwargs == {}


def test_call_overridden():
    provider = providers.Dict(a1="i1", a2="i2")
    overriding_provider1 = providers.Dict(a2="i2", a3="i3")
    overriding_provider2 = providers.Dict(a3="i3", a4="i4")

    provider.override(overriding_provider1)
    provider.override(overriding_provider2)

    instance1 = provider()
    instance2 = provider()

    assert instance1 is not instance2
    assert instance1 == {"a3": "i3", "a4": "i4"}
    assert instance2 == {"a3": "i3", "a4": "i4"}


def test_deepcopy():
    provider = providers.Dict(a1="i1", a2="i2")

    provider_copy = providers.deepcopy(provider)

    assert provider is not provider_copy
    assert provider.kwargs == provider_copy.kwargs
    assert isinstance(provider, providers.Dict)


def test_deepcopy_from_memo():
    provider = providers.Dict(a1="i1", a2="i2")
    provider_copy_memo = providers.Dict(a1="i1", a2="i2")

    provider_copy = providers.deepcopy(
        provider,
        memo={id(provider): provider_copy_memo},
    )

    assert provider_copy is provider_copy_memo


def test_deepcopy_kwargs():
    provider = providers.Dict()
    dependent_provider1 = providers.Factory(list)
    dependent_provider2 = providers.Factory(dict)

    provider.add_kwargs(d1=dependent_provider1, d2=dependent_provider2)

    provider_copy = providers.deepcopy(provider)
    dependent_provider_copy1 = provider_copy.kwargs["d1"]
    dependent_provider_copy2 = provider_copy.kwargs["d2"]

    assert provider.kwargs != provider_copy.kwargs

    assert dependent_provider1.cls is dependent_provider_copy1.cls
    assert dependent_provider1 is not dependent_provider_copy1

    assert dependent_provider2.cls is dependent_provider_copy2.cls
    assert dependent_provider2 is not dependent_provider_copy2


def test_deepcopy_kwargs_non_string_keys():
    a1 = object()
    a2 = object()

    dependent_provider1 = providers.Factory(list)
    dependent_provider2 = providers.Factory(dict)

    provider = providers.Dict({a1: dependent_provider1, a2: dependent_provider2})

    provider_copy = providers.deepcopy(provider)
    dependent_provider_copy1 = provider_copy.kwargs[a1]
    dependent_provider_copy2 = provider_copy.kwargs[a2]

    assert provider.kwargs != provider_copy.kwargs

    assert dependent_provider1.cls is dependent_provider_copy1.cls
    assert dependent_provider1 is not dependent_provider_copy1

    assert dependent_provider2.cls is dependent_provider_copy2.cls
    assert dependent_provider2 is not dependent_provider_copy2


def test_deepcopy_overridden():
    provider = providers.Dict()
    object_provider = providers.Object(object())

    provider.override(object_provider)

    provider_copy = providers.deepcopy(provider)
    object_provider_copy = provider_copy.overridden[0]

    assert provider is not provider_copy
    assert provider.kwargs == provider_copy.kwargs
    assert isinstance(provider, providers.Dict)

    assert object_provider is not object_provider_copy
    assert isinstance(object_provider_copy, providers.Object)


def test_deepcopy_with_sys_streams():
    provider = providers.Dict()
    provider.add_kwargs(stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)

    provider_copy = providers.deepcopy(provider)

    assert provider is not provider_copy
    assert isinstance(provider_copy, providers.Dict)
    assert provider.kwargs["stdin"] is sys.stdin
    assert provider.kwargs["stdout"] is sys.stdout
    assert provider.kwargs["stderr"] is sys.stderr


def test_repr():
    provider = providers.Dict(a1=1, a2=2)
    assert repr(provider) == (
        "<dependency_injector.providers."
        "Dict({0}) at {1}>".format(repr(provider.kwargs), hex(id(provider)))
    )



from typing import Any, Dict

from dependency_injector import providers


# Test 1: to check the return type (class)
provider1 = providers.Dict(
    a1=providers.Factory(object),
    a2=providers.Factory(object),
)
var1: Dict[Any, Any] = provider1()


# Test 2: to check init with non-string keys
provider2 = providers.Dict({object(): providers.Factory(object)})
var2: Dict[Any, Any] = provider2()


# Test 3: to check init with non-string keys
provider3 = providers.Dict({object(): providers.Factory(object)}, a2=providers.Factory(object))
var3: Dict[Any, Any] = provider3()


# Test 4: to check the .args attributes
provider4 = providers.Dict(
    a1=providers.Factory(object),
    a2=providers.Factory(object),
)
args4: Dict[Any, Any] = provider4.kwargs


# Test 5: to check the provided instance interface
provider5 = providers.Dict(
    a1=providers.Factory(object),
    a2=providers.Factory(object),
)
provided5: providers.ProvidedInstance = provider5.provided


# Test 6: to check the return type with await
provider6 = providers.Dict(
    a1=providers.Factory(object),
    a2=providers.Factory(object),
)
async def _async3() -> None:
    var1: Dict[Any, Any] = await provider6()  # type: ignore
    var2: Dict[Any, Any] = await provider6.async_()



from typing import Type, Optional

from dependency_injector import providers


# Test 1: to check the return type
provider1 = providers.Object(int(3))
var1: int = provider1()

# Test 2: to check the provided instance interface
provider2 = providers.Object(int)
provided2: providers.ProvidedInstance = provider2.provided
attr_getter2: providers.AttributeGetter = provider2.provided.attr
item_getter2: providers.ItemGetter = provider2.provided["item"]
method_caller2: providers.MethodCaller = provider2.provided.method.call(123, arg=324)

# Test 3: to check the return type with await
provider3 = providers.Object(int(3))
async def _async3() -> None:
    var1: int = await provider3()  # type: ignore
    var2: int = await provider3.async_()

# Test 4: to check class type from provider
provider4 = providers.Object(int("1"))
provided_provides: Optional[int] = provider4.provides



"""Resource provider async tests."""

import asyncio
import inspect
import sys
from typing import Any

from dependency_injector import containers, providers, resources
from pytest import mark, raises


@mark.asyncio
async def test_init_async_function():
    resource = object()

    async def _init():
        await asyncio.sleep(0.001)
        _init.counter += 1
        return resource

    _init.counter = 0

    provider = providers.Resource(_init)

    result1 = await provider()
    assert result1 is resource
    assert _init.counter == 1

    result2 = await provider()
    assert result2 is resource
    assert _init.counter == 1

    await provider.shutdown()


@mark.asyncio
@mark.skipif(sys.version_info < (3, 6), reason="requires Python 3.6+")
async def test_init_async_generator():
    resource = object()

    async def _init():
        await asyncio.sleep(0.001)
        _init.init_counter += 1

        yield resource

        await asyncio.sleep(0.001)
        _init.shutdown_counter += 1

    _init.init_counter = 0
    _init.shutdown_counter = 0

    provider = providers.Resource(_init)

    result1 = await provider()
    assert result1 is resource
    assert _init.init_counter == 1
    assert _init.shutdown_counter == 0

    await provider.shutdown()
    assert _init.init_counter == 1
    assert _init.shutdown_counter == 1

    result2 = await provider()
    assert result2 is resource
    assert _init.init_counter == 2
    assert _init.shutdown_counter == 1

    await provider.shutdown()
    assert _init.init_counter == 2
    assert _init.shutdown_counter == 2


@mark.asyncio
async def test_init_async_class():
    resource = object()

    class TestResource(resources.AsyncResource):
        init_counter = 0
        shutdown_counter = 0

        async def init(self):
            await asyncio.sleep(0.001)
            self.__class__.init_counter += 1
            return resource

        async def shutdown(self, resource_):
            await asyncio.sleep(0.001)
            self.__class__.shutdown_counter += 1
            assert resource_ is resource

    provider = providers.Resource(TestResource)

    result1 = await provider()
    assert result1 is resource
    assert TestResource.init_counter == 1
    assert TestResource.shutdown_counter == 0

    await provider.shutdown()
    assert TestResource.init_counter == 1
    assert TestResource.shutdown_counter == 1

    result2 = await provider()
    assert result2 is resource
    assert TestResource.init_counter == 2
    assert TestResource.shutdown_counter == 1

    await provider.shutdown()
    assert TestResource.init_counter == 2
    assert TestResource.shutdown_counter == 2


def test_init_async_class_generic_typing():
    # See issue: https://github.com/ets-labs/python-dependency-injector/issues/488
    class TestDependency:
        ...

    class TestAsyncResource(resources.AsyncResource[TestDependency]):
        async def init(self, *args: Any, **kwargs: Any) -> TestDependency:
            return TestDependency()

        async def shutdown(self, resource: TestDependency) -> None: ...

    assert issubclass(TestAsyncResource, resources.AsyncResource) is True


def test_init_async_class_abc_init_definition_is_required():
    class TestAsyncResource(resources.AsyncResource):
        ...

    with raises(TypeError) as context:
        TestAsyncResource()

    assert "Can't instantiate abstract class TestAsyncResource" in str(context.value)
    assert "init" in str(context.value)


def test_init_async_class_abc_shutdown_definition_is_not_required():
    class TestAsyncResource(resources.AsyncResource):
        async def init(self):
            ...

    assert hasattr(TestAsyncResource(), "shutdown") is True
    assert inspect.iscoroutinefunction(TestAsyncResource.shutdown) is True


@mark.asyncio
async def test_init_with_error():
    async def _init():
        raise RuntimeError()

    provider = providers.Resource(_init)

    future = provider()
    assert provider.initialized is True
    assert provider.is_async_mode_enabled() is True

    with raises(RuntimeError):
        await future

    assert provider.initialized is False
    assert provider.is_async_mode_enabled() is True


@mark.asyncio
async def test_init_async_gen_with_error():
    async def _init():
        raise RuntimeError()
        yield

    provider = providers.Resource(_init)

    future = provider()
    assert provider.initialized is True
    assert provider.is_async_mode_enabled() is True

    with raises(RuntimeError):
        await future

    assert provider.initialized is False
    assert provider.is_async_mode_enabled() is True


@mark.asyncio
async def test_init_async_subclass_with_error():
    class _Resource(resources.AsyncResource):
        async def init(self):
            raise RuntimeError()

        async def shutdown(self, resource):
            pass

    provider = providers.Resource(_Resource)

    future = provider()
    assert provider.initialized is True
    assert provider.is_async_mode_enabled() is True

    with raises(RuntimeError):
        await future

    assert provider.initialized is False
    assert provider.is_async_mode_enabled() is True


@mark.asyncio
async def test_init_with_dependency_to_other_resource():
    # See: https://github.com/ets-labs/python-dependency-injector/issues/361
    async def init_db_connection(db_url: str):
        await asyncio.sleep(0.001)
        yield {"connection": "OK", "url": db_url}

    async def init_user_session(db):
        await asyncio.sleep(0.001)
        yield {"session": "OK", "db": db}

    class Container(containers.DeclarativeContainer):
        config = providers.Configuration()

        db_connection = providers.Resource(
            init_db_connection,
            db_url=config.db_url,
        )

        user_session = providers.Resource(
            init_user_session,
            db=db_connection
        )

    async def main():
        container = Container(config={"db_url": "postgres://..."})
        try:
            return await container.user_session()
        finally:
            await container.shutdown_resources()

    result = await main()
    assert result == {"session": "OK", "db": {"connection": "OK", "url": "postgres://..."}}


@mark.asyncio
async def test_init_and_shutdown_methods():
    async def _init():
        await asyncio.sleep(0.001)
        _init.init_counter += 1

        yield

        await asyncio.sleep(0.001)
        _init.shutdown_counter += 1

    _init.init_counter = 0
    _init.shutdown_counter = 0

    provider = providers.Resource(_init)

    await provider.init()
    assert _init.init_counter == 1
    assert _init.shutdown_counter == 0

    await provider.shutdown()
    assert _init.init_counter == 1
    assert _init.shutdown_counter == 1

    await provider.init()
    assert _init.init_counter == 2
    assert _init.shutdown_counter == 1

    await provider.shutdown()
    assert _init.init_counter == 2
    assert _init.shutdown_counter == 2


@mark.asyncio
async def test_shutdown_of_not_initialized():
    async def _init():
        yield

    provider = providers.Resource(_init)
    provider.enable_async_mode()

    result = await provider.shutdown()
    assert result is None


@mark.asyncio
async def test_concurrent_init():
    resource = object()

    async def _init():
        await asyncio.sleep(0.001)
        _init.counter += 1
        return resource

    _init.counter = 0

    provider = providers.Resource(_init)

    result1, result2 = await asyncio.gather(
        provider(),
        provider()
    )

    assert result1 is resource
    assert _init.counter == 1

    assert result2 is resource
    assert _init.counter == 1



"""Factory provider tests."""

import decimal
import sys

from dependency_injector import providers, errors
from pytest import raises, mark

from .common import Example


def test_is_provider():
    assert providers.is_provider(providers.Factory(Example)) is True


def test_init_with_not_callable():
    with raises(errors.Error):
        providers.Factory(123)


def test_init_optional_provides():
    provider = providers.Factory()
    provider.set_provides(object)
    assert provider.provides is object
    assert isinstance(provider(), object)


def test_set_provides_returns_():
    provider = providers.Factory()
    assert provider.set_provides(object) is provider


@mark.parametrize(
    "str_name,cls",
    [
        ("dependency_injector.providers.Factory", providers.Factory),
        ("decimal.Decimal", decimal.Decimal),
        ("list", list),
        (".common.Example", Example),
        ("test_is_provider", test_is_provider),
    ],
)
def test_set_provides_string_imports(str_name, cls):
    assert providers.Factory(str_name).provides is cls


def test_init_with_valid_provided_type():
    class ExampleProvider(providers.Factory):
        provided_type = Example

    example_provider = ExampleProvider(Example, 1, 2)

    assert isinstance(example_provider(), Example)


def test_init_with_valid_provided_subtype():
    class ExampleProvider(providers.Factory):
        provided_type = Example

    class NewExample(Example):
        pass

    example_provider = ExampleProvider(NewExample, 1, 2)
    assert isinstance(example_provider(), NewExample)


def test_init_with_invalid_provided_type():
    class ExampleProvider(providers.Factory):
        provided_type = Example

    with raises(errors.Error):
        ExampleProvider(list)


def test_provided_instance_provider():
    provider = providers.Factory(Example)
    assert isinstance(provider.provided, providers.ProvidedInstance)


def test_call():
    provider = providers.Factory(Example)

    instance1 = provider()
    instance2 = provider()

    assert instance1 is not instance2
    assert isinstance(instance1, Example)
    assert isinstance(instance2, Example)


def test_call_with_init_positional_args():
    provider = providers.Factory(Example, "i1", "i2")

    instance1 = provider()
    instance2 = provider()

    assert instance1.init_arg1 == "i1"
    assert instance1.init_arg2 == "i2"

    assert instance2.init_arg1 == "i1"
    assert instance2.init_arg2 == "i2"

    assert instance1 is not instance2
    assert isinstance(instance1, Example)
    assert isinstance(instance2, Example)


def test_call_with_init_keyword_args():
    provider = providers.Factory(Example, init_arg1="i1", init_arg2="i2")

    instance1 = provider()
    instance2 = provider()

    assert instance1.init_arg1 == "i1"
    assert instance1.init_arg2 == "i2"

    assert instance2.init_arg1 == "i1"
    assert instance2.init_arg2 == "i2"

    assert instance1 is not instance2
    assert isinstance(instance1, Example)
    assert isinstance(instance2, Example)


def test_call_with_init_positional_and_keyword_args():
    provider = providers.Factory(Example, "i1", init_arg2="i2")

    instance1 = provider()
    instance2 = provider()

    assert instance1.init_arg1 == "i1"
    assert instance1.init_arg2 == "i2"

    assert instance2.init_arg1 == "i1"
    assert instance2.init_arg2 == "i2"

    assert instance1 is not instance2
    assert isinstance(instance1, Example)
    assert isinstance(instance2, Example)


def test_call_with_attributes():
    provider = providers.Factory(Example)
    provider.add_attributes(attribute1="a1", attribute2="a2")

    instance1 = provider()
    instance2 = provider()

    assert instance1.attribute1 == "a1"
    assert instance1.attribute2 == "a2"

    assert instance2.attribute1 == "a1"
    assert instance2.attribute2 == "a2"

    assert instance1 is not instance2
    assert isinstance(instance1, Example)
    assert isinstance(instance2, Example)


def test_call_with_context_args():
    provider = providers.Factory(Example, 11, 22)

    instance = provider(33, 44)

    assert instance.init_arg1 == 11
    assert instance.init_arg2 == 22
    assert instance.init_arg3 == 33
    assert instance.init_arg4 == 44


def test_call_with_context_kwargs():
    provider = providers.Factory(Example, init_arg1=1)

    instance1 = provider(init_arg2=22)
    assert instance1.init_arg1 == 1
    assert instance1.init_arg2 == 22

    instance2 = provider(init_arg1=11, init_arg2=22)
    assert instance2.init_arg1 == 11
    assert instance2.init_arg2 == 22


def test_call_with_context_args_and_kwargs():
    provider = providers.Factory(Example, 11)

    instance = provider(22, init_arg3=33, init_arg4=44)

    assert instance.init_arg1 == 11
    assert instance.init_arg2 == 22
    assert instance.init_arg3 == 33
    assert instance.init_arg4 == 44


def test_call_with_deep_context_kwargs():
    class Regularizer:
        def __init__(self, alpha):
            self.alpha = alpha

    class Loss:
        def __init__(self, regularizer):
            self.regularizer = regularizer

    class ClassificationTask:
        def __init__(self, loss):
            self.loss = loss

    class Algorithm:
        def __init__(self, task):
            self.task = task

    algorithm_factory = providers.Factory(
        Algorithm,
        task=providers.Factory(
            ClassificationTask,
            loss=providers.Factory(
                Loss,
                regularizer=providers.Factory(
                    Regularizer,
                ),
            ),
        ),
    )

    algorithm_1 = algorithm_factory(task__loss__regularizer__alpha=0.5)
    algorithm_2 = algorithm_factory(task__loss__regularizer__alpha=0.7)
    algorithm_3 = algorithm_factory(task__loss__regularizer=Regularizer(alpha=0.8))

    assert algorithm_1.task.loss.regularizer.alpha == 0.5
    assert algorithm_2.task.loss.regularizer.alpha == 0.7
    assert algorithm_3.task.loss.regularizer.alpha == 0.8


def test_fluent_interface():
    provider = providers.Factory(Example) \
        .add_args(1, 2) \
        .add_kwargs(init_arg3=3, init_arg4=4) \
        .add_attributes(attribute1=5, attribute2=6)

    instance = provider()

    assert instance.init_arg1 == 1
    assert instance.init_arg2 == 2
    assert instance.init_arg3 == 3
    assert instance.init_arg4 == 4
    assert instance.attribute1 == 5
    assert instance.attribute2 == 6


def test_set_args():
    provider = providers.Factory(Example) \
        .add_args(1, 2) \
        .set_args(3, 4)
    assert provider.args == (3, 4)


def test_set_kwargs():
    provider = providers.Factory(Example) \
        .add_kwargs(init_arg3=3, init_arg4=4) \
        .set_kwargs(init_arg3=4, init_arg4=5)
    assert provider.kwargs == dict(init_arg3=4, init_arg4=5)


def test_set_attributes():
    provider = providers.Factory(Example) \
        .add_attributes(attribute1=5, attribute2=6) \
        .set_attributes(attribute1=6, attribute2=7)
    assert provider.attributes == dict(attribute1=6, attribute2=7)


def test_clear_args():
    provider = providers.Factory(Example) \
        .add_args(1, 2) \
        .clear_args()
    assert provider.args == tuple()


def test_clear_kwargs():
    provider = providers.Factory(Example) \
        .add_kwargs(init_arg3=3, init_arg4=4) \
        .clear_kwargs()
    assert provider.kwargs == dict()


def test_clear_attributes():
    provider = providers.Factory(Example) \
        .add_attributes(attribute1=5, attribute2=6) \
        .clear_attributes()
    assert provider.attributes == dict()


def test_call_overridden():
    provider = providers.Factory(Example)
    overriding_provider1 = providers.Factory(dict)
    overriding_provider2 = providers.Factory(list)

    provider.override(overriding_provider1)
    provider.override(overriding_provider2)

    instance1 = provider()
    instance2 = provider()

    assert instance1 is not instance2
    assert isinstance(instance1, list)
    assert isinstance(instance2, list)


def test_deepcopy():
    provider = providers.Factory(Example)

    provider_copy = providers.deepcopy(provider)

    assert provider is not provider_copy
    assert provider.cls is provider_copy.cls
    assert isinstance(provider, providers.Factory)


def test_deepcopy_from_memo():
    provider = providers.Factory(Example)
    provider_copy_memo = providers.Factory(Example)

    provider_copy = providers.deepcopy(provider, memo={id(provider): provider_copy_memo})

    assert provider_copy is provider_copy_memo


def test_deepcopy_args():
    provider = providers.Factory(Example)
    dependent_provider1 = providers.Factory(list)
    dependent_provider2 = providers.Factory(dict)

    provider.add_args(dependent_provider1, dependent_provider2)

    provider_copy = providers.deepcopy(provider)
    dependent_provider_copy1 = provider_copy.args[0]
    dependent_provider_copy2 = provider_copy.args[1]

    assert provider.args != provider_copy.args

    assert dependent_provider1.cls is dependent_provider_copy1.cls
    assert dependent_provider1 is not dependent_provider_copy1

    assert dependent_provider2.cls is dependent_provider_copy2.cls
    assert dependent_provider2 is not dependent_provider_copy2


def test_deepcopy_kwargs():
    provider = providers.Factory(Example)
    dependent_provider1 = providers.Factory(list)
    dependent_provider2 = providers.Factory(dict)

    provider.add_kwargs(a1=dependent_provider1, a2=dependent_provider2)

    provider_copy = providers.deepcopy(provider)
    dependent_provider_copy1 = provider_copy.kwargs["a1"]
    dependent_provider_copy2 = provider_copy.kwargs["a2"]

    assert provider.kwargs != provider_copy.kwargs

    assert dependent_provider1.cls is dependent_provider_copy1.cls
    assert dependent_provider1 is not dependent_provider_copy1

    assert dependent_provider2.cls is dependent_provider_copy2.cls
    assert dependent_provider2 is not dependent_provider_copy2


def test_deepcopy_attributes():
    provider = providers.Factory(Example)
    dependent_provider1 = providers.Factory(list)
    dependent_provider2 = providers.Factory(dict)

    provider.add_attributes(a1=dependent_provider1, a2=dependent_provider2)

    provider_copy = providers.deepcopy(provider)
    dependent_provider_copy1 = provider_copy.attributes["a1"]
    dependent_provider_copy2 = provider_copy.attributes["a2"]

    assert provider.attributes != provider_copy.attributes

    assert dependent_provider1.cls is dependent_provider_copy1.cls
    assert dependent_provider1 is not dependent_provider_copy1

    assert dependent_provider2.cls is dependent_provider_copy2.cls
    assert dependent_provider2 is not dependent_provider_copy2


def test_deepcopy_overridden():
    provider = providers.Factory(Example)
    object_provider = providers.Object(object())

    provider.override(object_provider)

    provider_copy = providers.deepcopy(provider)
    object_provider_copy = provider_copy.overridden[0]

    assert provider is not provider_copy
    assert provider.cls is provider_copy.cls
    assert isinstance(provider, providers.Factory)

    assert object_provider is not object_provider_copy
    assert isinstance(object_provider_copy, providers.Object)


def test_deepcopy_with_sys_streams():
    provider = providers.Factory(Example)
    provider.add_args(sys.stdin)
    provider.add_kwargs(a2=sys.stdout)
    provider.add_attributes(a3=sys.stderr)

    provider_copy = providers.deepcopy(provider)

    assert provider is not provider_copy
    assert isinstance(provider_copy, providers.Factory)
    assert provider.args[0] is sys.stdin
    assert provider.kwargs["a2"] is sys.stdout
    assert provider.attributes["a3"] is sys.stderr


def test_repr():
    provider = providers.Factory(Example)
    assert repr(provider) == (
        "<dependency_injector.providers."
        "Factory({0}) at {1}>".format(repr(Example), hex(id(provider)))
    )



"""Test providers performance."""

import time
import gc

import dependency_injector.providers


class Tester(object):
    """Performance tester for provider module implementations."""

    def __init__(self, provider_modules, duration_factor):
        """Initializer."""
        self.provider_modules = provider_modules
        self.tests = [getattr(self, name)
                      for name in dir(self)
                      if name.startswith("test")]
        self.total_time = 0
        self.duration_factor = duration_factor

    def run(self):
        """Run all tests for all provider modules."""
        for module in self.provider_modules:
            print("\n")
            print("Running tests for module - \"{module}\":"
                  .format(module=module.__name__))

            gc.disable()
            for test in self.tests:
                start_time = time.time()
                test(module)
                self.total_time = time.time() - start_time
                print("Test \"{test}\" took - {time}"
                      .format(test=test.__name__,
                              time=self.total_time))
                gc.collect()

        gc.enable()
        print("\n")

    def test_raw_3_kw_injections(self, providers):
        """Test 3 keyword argument injections."""
        class A(object):
            pass

        class B(object):
            pass

        class C(object):
            pass

        class Test(object):
            def __init__(self, a, b, c):
                pass

        for x in range(int(5000000 * self.duration_factor)):
            Test(a=A(), b=B(), c=C())

    def test_factory_3_factory_kw_injections(self, providers):
        """Test factory with 3 keyword argument injections via factories."""
        class A(object):
            pass

        class B(object):
            pass

        class C(object):
            pass

        class Test(object):
            def __init__(self, a, b, c):
                pass

        a_factory = providers.Factory(A)
        b_factory = providers.Factory(B)
        c_factory = providers.Factory(C)
        test_factory = providers.Factory(Test,
                                         a=a_factory,
                                         b=b_factory,
                                         c=c_factory)
        for x in range(int(5000000 * self.duration_factor)):
            test_factory()

    def test_abstract_factory_3_factory_kw_injections(self, providers):
        """Test factory with 3 keyword argument injections via factories."""
        class A(object):
            pass

        class B(object):
            pass

        class C(object):
            pass

        class Test(object):
            def __init__(self, a, b, c):
                pass

        a_factory = providers.Factory(A)
        b_factory = providers.Factory(B)
        c_factory = providers.Factory(C)
        test_factory = providers.AbstractFactory(object)
        test_factory.override(providers.Factory(Test,
                                                a=a_factory,
                                                b=b_factory,
                                                c=c_factory))
        for x in range(int(5000000 * self.duration_factor)):
            test_factory()

    def test_factory_6_factory_kw_injections_0_context(self, providers):
        """Test factory with 6 keyword argument injections."""
        class Test(object):
            def __init__(self, a, b, c, d, e, f):
                pass

        test_factory = providers.Factory(Test, a=1, b=2, c=3, d=4, e=5, f=6)
        for x in range(int(5000000 * self.duration_factor)):
            test_factory()

    def test_factory_6_factory_kw_injections_1_context(self, providers):
        """Test factory with 6 keyword argument injections."""
        class Test(object):
            def __init__(self, a, b, c, d, e, f):
                pass

        test_factory = providers.Factory(Test, f=6)
        for x in range(int(5000000 * self.duration_factor)):
            test_factory(a=1, b=2, c=3, d=4, e=5)

    def test_factory_6_factory_kw_injections_3_context(self, providers):
        """Test factory with 6 keyword argument injections."""
        class Test(object):
            def __init__(self, a, b, c, d, e, f):
                pass

        test_factory = providers.Factory(Test, a=1, b=2, c=3)
        for x in range(int(5000000 * self.duration_factor)):
            test_factory(d=4, e=5, f=6)


if __name__ == "__main__":
    tester = Tester(
        provider_modules=[
            dependency_injector.providers,
        ],
        duration_factor=0.5)
    tester.run()



"""Main container instance tests."""

from dependency_injector import containers, providers, errors
from pytest import raises


class Container(containers.DeclarativeContainer):
    p11 = providers.Provider()
    p12 = providers.Provider()


def test_providers_attribute():
    container_1 = Container()
    container_2 = Container()

    assert container_1.p11 is not container_2.p11
    assert container_1.p12 is not container_2.p12
    assert container_1.providers != container_2.providers


def test_dependencies_attribute():
    container = Container()
    container.a1 = providers.Dependency()
    container.a2 = providers.DependenciesContainer()
    assert container.dependencies == {"a1": container.a1, "a2": container.a2}


def test_set_get_del_providers():
    p13 = providers.Provider()

    container_1 = Container()
    container_2 = Container()

    container_1.p13 = p13
    container_2.p13 = p13

    assert Container.providers == dict(p11=Container.p11, p12=Container.p12)
    assert Container.cls_providers, dict(p11=Container.p11, p12=Container.p12)

    assert container_1.providers == dict(p11=container_1.p11, p12=container_1.p12, p13=p13)
    assert container_2.providers == dict(p11=container_2.p11, p12=container_2.p12, p13=p13)

    del container_1.p13
    assert container_1.providers == dict(p11=container_1.p11, p12=container_1.p12)

    del container_2.p13
    assert container_2.providers == dict(p11=container_2.p11, p12=container_2.p12)

    del container_1.p11
    del container_1.p12
    assert container_1.providers == dict()
    assert Container.providers == dict(p11=Container.p11, p12=Container.p12)

    del container_2.p11
    del container_2.p12
    assert container_2.providers == dict()
    assert Container.providers == dict(p11=Container.p11, p12=Container.p12)


def test_set_invalid_provider_type():
    container = Container()
    container.provider_type = providers.Object

    with raises(errors.Error):
        container.px = providers.Provider()

    assert Container.provider_type is containers.DeclarativeContainer.provider_type


def test_set_providers():
    p13 = providers.Provider()
    p14 = providers.Provider()
    container = Container()

    container.set_providers(p13=p13, p14=p14)

    assert container.p13 is p13
    assert container.p14 is p14


def test_override():
    class _Container(containers.DeclarativeContainer):
        p11 = providers.Provider()

    class _OverridingContainer1(containers.DeclarativeContainer):
        p11 = providers.Provider()

    class _OverridingContainer2(containers.DeclarativeContainer):
        p11 = providers.Provider()
        p12 = providers.Provider()

    container = _Container()
    overriding_container1 = _OverridingContainer1()
    overriding_container2 = _OverridingContainer2()

    container.override(overriding_container1)
    container.override(overriding_container2)

    assert container.overridden == (overriding_container1, overriding_container2)
    assert container.p11.overridden == (overriding_container1.p11, overriding_container2.p11)

    assert _Container.overridden == tuple()
    assert _Container.p11.overridden == tuple()


def test_override_with_it():
    container = Container()
    with raises(errors.Error):
        container.override(container)


def test_override_providers():
    p1 = providers.Provider()
    p2 = providers.Provider()
    container = Container()

    container.override_providers(p11=p1, p12=p2)

    assert container.p11.last_overriding is p1
    assert container.p12.last_overriding is p2


def test_override_providers_context_manager():
    p1 = providers.Provider()
    p2 = providers.Provider()
    container = Container()

    with container.override_providers(p11=p1, p12=p2) as context_container:
        assert container is context_container
        assert container.p11.last_overriding is p1
        assert container.p12.last_overriding is p2

    assert container.p11.last_overriding is None
    assert container.p12.last_overriding is None


def test_override_providers_with_unknown_provider():
    container = Container()
    with raises(AttributeError):
        container.override_providers(unknown=providers.Provider())


def test_reset_last_overriding():
    class _Container(containers.DeclarativeContainer):
        p11 = providers.Provider()

    class _OverridingContainer1(containers.DeclarativeContainer):
        p11 = providers.Provider()

    class _OverridingContainer2(containers.DeclarativeContainer):
        p11 = providers.Provider()
        p12 = providers.Provider()

    container = _Container()
    overriding_container1 = _OverridingContainer1()
    overriding_container2 = _OverridingContainer2()

    container.override(overriding_container1)
    container.override(overriding_container2)
    container.reset_last_overriding()

    assert container.overridden == (overriding_container1,)
    assert container.p11.overridden, (overriding_container1.p11,)


def test_reset_last_overriding_when_not_overridden():
    container = Container()
    with raises(errors.Error):
        container.reset_last_overriding()


def test_reset_override():
    class _Container(containers.DeclarativeContainer):
        p11 = providers.Provider()

    class _OverridingContainer1(containers.DeclarativeContainer):
        p11 = providers.Provider()

    class _OverridingContainer2(containers.DeclarativeContainer):
        p11 = providers.Provider()
        p12 = providers.Provider()

    container = _Container()
    overriding_container1 = _OverridingContainer1()
    overriding_container2 = _OverridingContainer2()

    container.override(overriding_container1)
    container.override(overriding_container2)
    container.reset_override()

    assert container.overridden == tuple()
    assert container.p11.overridden == tuple()


def test_init_and_shutdown_resources_ordering():
    """Test init and shutdown resources.

    Methods .init_resources() and .shutdown_resources() should respect resources dependencies.
    Initialization should first initialize resources without dependencies and then provide
    these resources to other resources. Resources shutdown should follow the same rule: first
    shutdown resources without initialized dependencies and then continue correspondingly
    until all resources are shutdown.
    """
    initialized_resources = []
    shutdown_resources = []

    def _resource(name, **_):
        initialized_resources.append(name)
        yield name
        shutdown_resources.append(name)

    class Container(containers.DeclarativeContainer):
        resource1 = providers.Resource(
            _resource,
            name="r1",
        )
        resource2 = providers.Resource(
            _resource,
            name="r2",
            r1=resource1,
        )
        resource3 = providers.Resource(
            _resource,
            name="r3",
            r2=resource2,
        )

    container = Container()

    container.init_resources()
    assert initialized_resources == ["r1", "r2", "r3"]
    assert shutdown_resources == []

    container.shutdown_resources()
    assert initialized_resources == ["r1", "r2", "r3"]
    assert shutdown_resources == ["r3", "r2", "r1"]

    container.init_resources()
    assert initialized_resources == ["r1", "r2", "r3", "r1", "r2", "r3"]
    assert shutdown_resources == ["r3", "r2", "r1"]

    container.shutdown_resources()
    assert initialized_resources == ["r1", "r2", "r3", "r1", "r2", "r3"]
    assert shutdown_resources == ["r3", "r2", "r1", "r3", "r2", "r1"]


def test_shutdown_resources_circular_dependencies_breaker():
    def _resource(name, **_):
        yield name

    class Container(containers.DeclarativeContainer):
        resource1 = providers.Resource(
            _resource,
            name="r1",
        )
        resource2 = providers.Resource(
            _resource,
            name="r2",
            r1=resource1,
        )
        resource3 = providers.Resource(
            _resource,
            name="r3",
            r2=resource2,
        )

    container = Container()
    container.init_resources()

    # Create circular dependency after initialization (r3 -> r2 -> r1 -> r3 -> ...)
    container.resource1.add_kwargs(r3=container.resource3)

    with raises(RuntimeError, match="Unable to resolve resources shutdown order"):
        container.shutdown_resources()


def test_init_shutdown_nested_resources():
    def _init1():
        _init1.init_counter += 1
        yield
        _init1.shutdown_counter += 1

    _init1.init_counter = 0
    _init1.shutdown_counter = 0

    def _init2():
        _init2.init_counter += 1
        yield
        _init2.shutdown_counter += 1

    _init2.init_counter = 0
    _init2.shutdown_counter = 0

    class Container(containers.DeclarativeContainer):

        service = providers.Factory(
            dict,
            resource1=providers.Resource(_init1),
            resource2=providers.Resource(_init2),
        )

    container = Container()
    assert _init1.init_counter == 0
    assert _init1.shutdown_counter == 0
    assert _init2.init_counter == 0
    assert _init2.shutdown_counter == 0

    container.init_resources()
    assert _init1.init_counter == 1
    assert _init1.shutdown_counter == 0
    assert _init2.init_counter == 1
    assert _init2.shutdown_counter == 0

    container.shutdown_resources()
    assert _init1.init_counter == 1
    assert _init1.shutdown_counter == 1
    assert _init2.init_counter == 1
    assert _init2.shutdown_counter == 1

    container.init_resources()
    container.shutdown_resources()
    assert _init1.init_counter == 2
    assert _init1.shutdown_counter == 2
    assert _init2.init_counter == 2
    assert _init2.shutdown_counter == 2


def test_reset_singletons():
    class SubSubContainer(containers.DeclarativeContainer):
        singleton = providers.Singleton(object)

    class SubContainer(containers.DeclarativeContainer):
        singleton = providers.Singleton(object)
        sub_sub_container = providers.Container(SubSubContainer)

    class Container(containers.DeclarativeContainer):
        singleton = providers.Singleton(object)
        sub_container = providers.Container(SubContainer)

    container = Container()

    obj11 = container.singleton()
    obj12 = container.sub_container().singleton()
    obj13 = container.sub_container().sub_sub_container().singleton()

    obj21 = container.singleton()
    obj22 = container.sub_container().singleton()
    obj23 = container.sub_container().sub_sub_container().singleton()

    assert obj11 is obj21
    assert obj12 is obj22
    assert obj13 is obj23

    container.reset_singletons()

    obj31 = container.singleton()
    obj32 = container.sub_container().singleton()
    obj33 = container.sub_container().sub_sub_container().singleton()

    obj41 = container.singleton()
    obj42 = container.sub_container().singleton()
    obj43 = container.sub_container().sub_sub_container().singleton()

    assert obj11 is not obj31
    assert obj12 is not obj32
    assert obj13 is not obj33

    assert obj21 is not obj31
    assert obj22 is not obj32
    assert obj23 is not obj33

    assert obj31 is obj41
    assert obj32 is obj42
    assert obj33 is obj43


def test_reset_singletons_context_manager():
    class Item:
        def __init__(self, dependency):
            self.dependency = dependency

    class Container(containers.DeclarativeContainer):
        dependent = providers.Singleton(object)
        singleton = providers.Singleton(Item, dependency=dependent)

    container = Container()

    instance1 = container.singleton()
    with container.reset_singletons():
        instance2 = container.singleton()
    instance3 = container.singleton()

    assert len({instance1, instance2, instance3}) == 3
    assert len({instance1.dependency, instance2.dependency, instance3.dependency}) == 3


def test_reset_singletons_context_manager_as_attribute():
    container = containers.DeclarativeContainer()
    with container.reset_singletons() as alias:
        pass
    assert container is alias


def test_check_dependencies():
    class SubContainer(containers.DeclarativeContainer):
        dependency = providers.Dependency()

    class Container(containers.DeclarativeContainer):
        dependency = providers.Dependency()
        dependencies_container = providers.DependenciesContainer()
        provider = providers.List(dependencies_container.dependency)
        sub_container = providers.Container(SubContainer)

    container = Container()

    with raises(errors.Error) as exception_info:
        container.check_dependencies()

    assert "Container \"Container\" has undefined dependencies:" in str(exception_info.value)
    assert "\"Container.dependency\"" in str(exception_info.value)
    assert "\"Container.dependencies_container.dependency\"" in str(exception_info.value)
    assert "\"Container.sub_container.dependency\"" in str(exception_info.value)


def test_check_dependencies_all_defined():
    class Container(containers.DeclarativeContainer):
        dependency = providers.Dependency()

    container = Container(dependency="provided")
    result = container.check_dependencies()

    assert result is None


def test_assign_parent():
    parent = providers.DependenciesContainer()
    container = Container()

    container.assign_parent(parent)

    assert container.parent is parent


def test_parent_name_declarative_parent():
    container = Container()
    assert container.parent_name == "Container"


def test_parent_name():
    container = Container()
    assert container.parent_name == "Container"


def test_parent_name_with_deep_parenting():
    class Container2(containers.DeclarativeContainer):
        name = providers.Container(Container)

    class Container1(containers.DeclarativeContainer):
        container = providers.Container(Container2)

    container = Container1()
    assert container.container().name.parent_name == "Container1.container.name"


def test_parent_name_is_none():
    container = containers.DynamicContainer()
    assert container.parent_name is None


def test_parent_deepcopy():
    class ParentContainer(containers.DeclarativeContainer):
        child = providers.Container(Container)

    container = ParentContainer()
    copied = providers.deepcopy(container)

    assert container.child.parent is container
    assert copied.child.parent is copied

    assert container is not copied
    assert container.child is not copied.child
    assert container.child.parent is not copied.child.parent


def test_resolve_provider_name():
    container = Container()
    assert container.resolve_provider_name(container.p11) == "p11"


def test_resolve_provider_name_no_provider():
    container = Container()
    with raises(errors.Error):
        container.resolve_provider_name(providers.Provider())



from dependency_injector.wiring import inject, Provide

from ...container import Container
from ...service import Service


@inject
def test_function(service: Service = Provide[Container.service]):
    return service



from typing import Dict

from dependency_injector import containers, providers


# Test 1: to check declarative container subclass
class Container1(containers.DeclarativeContainer):
    provider = providers.Factory(int)


container1 = Container1()
container1_type: containers.Container = Container1()
provider1: providers.Provider = container1.provider
val1: int = container1.provider(3)


# Test 2: to check @override decorator
class Container21(containers.DeclarativeContainer):
    provider = providers.Factory(int)


@containers.override(Container21)
class Container22(containers.DeclarativeContainer):
    ...


# Test 3: to check @copy decorator
class Container31(containers.DeclarativeContainer):
    provider = providers.Factory(int)


@containers.copy(Container31)
class Container32(containers.DeclarativeContainer):
    ...


# Test 4: to override()
class Container4(containers.DeclarativeContainer):
    provider = providers.Factory(int)

container4 = Container4()
container4.override(Container4())


# Test 5: to check .dependencies attribute
class Container5(containers.DeclarativeContainer):
    provider = providers.Factory(int)


dependencies: Dict[str, providers.Provider] = Container5.dependencies


# Test 6: to check base class
class Container6(containers.DeclarativeContainer):
    provider = providers.Factory(int)


container6: containers.Container = Container6()


# Test 7: to override_providers()
class Container7(containers.DeclarativeContainer):
    provider = providers.Factory(str)

container7 = Container7()
container7.override_providers(provider="new_value")
with container7.override_providers(a=providers.Provider()):
    ...



"""Services module."""

import logging
import sqlite3
from typing import Dict

from mypy_boto3_s3 import S3Client


class BaseService:

    def __init__(self) -> None:
        self.logger = logging.getLogger(
            f"{__name__}.{self.__class__.__name__}",
        )


class UserService(BaseService):

    def __init__(self, db: sqlite3.Connection) -> None:
        self.db = db
        super().__init__()

    def get_user(self, email: str) -> Dict[str, str]:
        self.logger.debug("User %s has been found in database", email)
        return {"email": email, "password_hash": "..."}


class AuthService(BaseService):

    def __init__(self, db: sqlite3.Connection, token_ttl: int) -> None:
        self.db = db
        self.token_ttl = token_ttl
        super().__init__()

    def authenticate(self, user: Dict[str, str], password: str) -> None:
        assert password is not None
        self.logger.debug(
            "User %s has been successfully authenticated",
            user["email"],
        )


class PhotoService(BaseService):

    def __init__(self, db: sqlite3.Connection, s3: S3Client) -> None:
        self.db = db
        self.s3 = s3
        super().__init__()

    def upload_photo(self, user: Dict[str, str], photo_path: str) -> None:
        self.logger.debug(
            "Photo %s has been successfully uploaded by user %s",
            photo_path,
            user["email"],
        )



from dependency_injector import providers


# Test 1: to check the getattr type
provider1 = providers.DependenciesContainer(
    a=providers.Provider(),
    b=providers.Provider(),
)
a1: providers.Provider = provider1.a
b1: providers.Provider = provider1.b
c1: providers.ProvidedInstance = provider1.c.provided



from pathlib import Path
from typing import (
    Generic,
    Type,
    Dict,
    List,
    Tuple,
    Optional,
    Any,
    Union,
    ClassVar,
    Callable as _Callable,
    Iterable,
    Iterator,
    TypeVar,
    Awaitable,
    overload,
)

from .providers import Provider, Self, ProviderParent


C_Base = TypeVar("C_Base", bound="Container")
C = TypeVar("C", bound="DeclarativeContainer")
C_Overriding = TypeVar("C_Overriding", bound="DeclarativeContainer")
T = TypeVar("T")
TT = TypeVar("TT")


class WiringConfiguration:
    modules: List[Any]
    packages: List[Any]
    from_package: Optional[str]
    auto_wire: bool
    def __init__(self, modules: Optional[Iterable[Any]] = None, packages: Optional[Iterable[Any]] = None, from_package: Optional[str] = None, auto_wire: bool = True) -> None: ...


class Container:
    provider_type: Type[Provider] = Provider
    providers: Dict[str, Provider]
    dependencies: Dict[str, Provider]
    overridden: Tuple[Provider]
    wiring_config: WiringConfiguration
    auto_load_config: bool = True
    __self__: Self
    def __init__(self) -> None: ...
    def __deepcopy__(self, memo: Optional[Dict[str, Any]]) -> Provider: ...
    def __setattr__(self, name: str, value: Union[Provider, Any]) -> None: ...
    def __getattr__(self, name: str) -> Provider: ...
    def __delattr__(self, name: str) -> None: ...
    def set_providers(self, **providers: Provider): ...
    def set_provider(self, name: str, provider: Provider) -> None: ...
    def override(self, overriding: Union[Container, Type[Container]]) -> None: ...
    def override_providers(self, **overriding_providers: Union[Provider, Any]) -> ProvidersOverridingContext[C_Base]: ...
    def reset_last_overriding(self) -> None: ...
    def reset_override(self) -> None: ...
    def is_auto_wiring_enabled(self) -> bool: ...
    def wire(self, modules: Optional[Iterable[Any]] = None, packages: Optional[Iterable[Any]] = None, from_package: Optional[str] = None) -> None: ...
    def unwire(self) -> None: ...
    def init_resources(self) -> Optional[Awaitable]: ...
    def shutdown_resources(self) -> Optional[Awaitable]: ...
    def load_config(self) -> None: ...
    def apply_container_providers_overridings(self) -> None: ...
    def reset_singletons(self) -> SingletonResetContext[C_Base]: ...
    def check_dependencies(self) -> None: ...
    def from_schema(self, schema: Dict[Any, Any]) -> None: ...
    def from_yaml_schema(self, filepath: Union[Path, str], loader: Optional[Any]=None) -> None: ...
    def from_json_schema(self, filepath: Union[Path, str]) -> None: ...
    @overload
    def resolve_provider_name(self, provider: Provider) -> str: ...
    @classmethod
    @overload
    def resolve_provider_name(cls, provider: Provider) -> str: ...
    @property
    def parent(self) -> Optional[ProviderParent]: ...
    @property
    def parent_name(self) -> Optional[str]: ...
    def assign_parent(self, parent: ProviderParent) -> None: ...
    @overload
    def traverse(self, types: Optional[Iterable[Type[TT]]] = None) -> Iterator[TT]: ...
    @classmethod
    @overload
    def traverse(cls, types: Optional[Iterable[Type[TT]]] = None) -> Iterator[TT]: ...


class DynamicContainer(Container): ...


class DeclarativeContainer(Container):
    cls_providers: ClassVar[Dict[str, Provider]]
    inherited_providers: ClassVar[Dict[str, Provider]]
    def __init__(self, **overriding_providers: Union[Provider, Any]) -> None: ...
    @classmethod
    def override(cls, overriding: Union[Container, Type[Container]]) -> None: ...
    @classmethod
    def override_providers(cls, **overriding_providers: Union[Provider, Any]) -> ProvidersOverridingContext[C_Base]: ...
    @classmethod
    def reset_last_overriding(cls) -> None: ...
    @classmethod
    def reset_override(cls) -> None: ...


class ProvidersOverridingContext(Generic[T]):
    def __init__(self, container: T, overridden_providers: Iterable[Union[Provider, Any]]) -> None: ...
    def __enter__(self) -> T: ...
    def __exit__(self, *_: Any) -> None: ...


class SingletonResetContext(Generic[T]):
    def __init__(self, container: T): ...
    def __enter__(self) -> T: ...
    def __exit__(self, *_: Any) -> None: ...


def override(container: Type[C]) -> _Callable[[Type[C_Overriding]], Type[C_Overriding]]: ...


def copy(container: Type[C]) -> _Callable[[Type[C_Overriding]], Type[C_Overriding]]: ...


def is_container(instance: Any) -> bool: ...



"""Object provider traversal tests."""

from dependency_injector import providers


def test_traversal():
    provider = providers.Object("string")
    all_providers = list(provider.traverse())
    assert len(all_providers) == 0


def test_traversal_provider():
    another_provider = providers.Provider()
    provider = providers.Object(another_provider)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 1
    assert another_provider in all_providers


def test_traversal_provider_and_overriding():
    another_provider_1 = providers.Provider()
    another_provider_2 = providers.Provider()
    another_provider_3 = providers.Provider()

    provider = providers.Object(another_provider_1)

    provider.override(another_provider_2)
    provider.override(another_provider_3)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 3
    assert another_provider_1 in all_providers
    assert another_provider_2 in all_providers
    assert another_provider_3 in all_providers



"""CallableDelegate provider tests."""

from dependency_injector import providers, errors
from pytest import raises

from .common import example


def test_is_delegate():
    provider = providers.Callable(example)
    delegate = providers.CallableDelegate(provider)
    assert isinstance(delegate, providers.Delegate)


def test_init_with_not_callable():
    with raises(errors.Error):
        providers.CallableDelegate(providers.Object(object()))



"""Configuration.from_dict() tests."""

from pytest import mark, raises


CONFIG_OPTIONS_1 = {
    "section1": {
        "value1": "1",
    },
    "section2": {
        "value2": "2",
    },
}
CONFIG_OPTIONS_2 = {
    "section1": {
        "value1": "11",
        "value11": "11",
    },
    "section3": {
        "value3": "3",
    },
}


def test(config):
    config.from_dict(CONFIG_OPTIONS_1)

    assert config() == {"section1": {"value1": "1"}, "section2": {"value2": "2"}}
    assert config.section1() == {"value1": "1"}
    assert config.section1.value1() == "1"
    assert config.section2() == {"value2": "2"}
    assert config.section2.value2() == "2"


def test_merge(config):
    config.from_dict(CONFIG_OPTIONS_1)
    config.from_dict(CONFIG_OPTIONS_2)

    assert config() == {
        "section1": {
            "value1": "11",
            "value11": "11",
        },
        "section2": {
            "value2": "2",
        },
        "section3": {
            "value3": "3",
        },
    }
    assert config.section1() == {"value1": "11", "value11": "11"}
    assert config.section1.value1() == "11"
    assert config.section1.value11() == "11"
    assert config.section2() == {"value2": "2"}
    assert config.section2.value2() == "2"
    assert config.section3() == {"value3": "3"}
    assert config.section3.value3() == "3"


def test_empty_dict(config):
    config.from_dict({})
    assert config() == {}


def test_option_empty_dict(config):
    config.option.from_dict({})
    assert config.option() == {}


@mark.parametrize("config_type", ["strict"])
def test_empty_dict_in_strict_mode(config):
    with raises(ValueError):
        config.from_dict({})


@mark.parametrize("config_type", ["strict"])
def test_option_empty_dict_in_strict_mode(config):
    with raises(ValueError):
        config.option.from_dict({})


def test_required_empty_dict(config):
    with raises(ValueError):
        config.from_dict({}, required=True)


def test_required_option_empty_dict(config):
    with raises(ValueError):
        config.option.from_dict({}, required=True)


@mark.parametrize("config_type", ["strict"])
def test_not_required_empty_dict_strict_mode(config):
    config.from_dict({}, required=False)
    assert config() == {}


@mark.parametrize("config_type", ["strict"])
def test_not_required_option_empty_dict_strict_mode(config):
    config.option.from_dict({}, required=False)
    assert config.option() == {}
    assert config() == {"option": {}}




from dependency_injector import providers


class Animal:
    ...


class Cat(Animal):
    ...


# Test 1: to check Aggregate provider
provider1: providers.Aggregate[str] = providers.Aggregate(
    a=providers.Object("str1"),
    b=providers.Object("str2"),
)
provider_a_1: providers.Provider[str] = provider1.a
provider_b_1: providers.Provider[str] = provider1.b
val1: str = provider1("a")

provider1_set_non_string_keys: providers.Aggregate[str] = providers.Aggregate()
provider1_set_non_string_keys.set_providers({Cat: providers.Object("str")})
provider_set_non_string_1: providers.Provider[str] = provider1_set_non_string_keys.providers[Cat]

provider1_new_non_string_keys: providers.Aggregate[str] = providers.Aggregate(
    {Cat: providers.Object("str")},
)
factory_new_non_string_1: providers.Provider[str] = provider1_new_non_string_keys.providers[Cat]

provider1_no_explicit_typing = providers.Aggregate(a=providers.Object("str"))
provider1_no_explicit_typing_factory: providers.Provider[str] = provider1_no_explicit_typing.providers["a"]
provider1_no_explicit_typing_object: str = provider1_no_explicit_typing("a")



"""Dict provider traversal tests."""

from dependency_injector import providers


def test_traverse_kwargs():
    provider1 = providers.Object("bar")
    provider2 = providers.Object("baz")
    provider = providers.Dict(foo="foo", bar=provider1, baz=provider2)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 2
    assert provider1 in all_providers
    assert provider2 in all_providers


def test_traverse_overridden():
    provider1 = providers.Object("bar")
    provider2 = providers.Object("baz")
    provider3 = providers.Dict(bar=provider1, baz=provider2)

    provider = providers.Dict(foo="foo")
    provider.override(provider3)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 3
    assert provider1 in all_providers
    assert provider2 in all_providers
    assert provider3 in all_providers



"""Configuration.from_ini() tests."""

from dependency_injector import errors
from pytest import mark, raises


def test(config, ini_config_file_1):
    config.from_ini(ini_config_file_1)

    assert config() == {"section1": {"value1": "1"}, "section2": {"value2": "2"}}
    assert config.section1() == {"value1": "1"}
    assert config.section1.value1() == "1"
    assert config.section2() == {"value2": "2"}
    assert config.section2.value2() == "2"


def test_option(config, ini_config_file_1):
    config.option.from_ini(ini_config_file_1)

    assert config() == {"option": {"section1": {"value1": "1"}, "section2": {"value2": "2"}}}
    assert config.option() == {"section1": {"value1": "1"}, "section2": {"value2": "2"}}
    assert config.option.section1() == {"value1": "1"}
    assert config.option.section1.value1() == "1"
    assert config.option.section2() == {"value2": "2"}
    assert config.option.section2.value2() == "2"


def test_merge(config, ini_config_file_1, ini_config_file_2):
    config.from_ini(ini_config_file_1)
    config.from_ini(ini_config_file_2)

    assert config() == {
        "section1": {
            "value1": "11",
            "value11": "11",
        },
        "section2": {
            "value2": "2",
        },
        "section3": {
            "value3": "3",
        },
    }
    assert config.section1() == {"value1": "11", "value11": "11"}
    assert config.section1.value1() == "11"
    assert config.section1.value11() == "11"
    assert config.section2() == {"value2": "2"}
    assert config.section2.value2() == "2"
    assert config.section3() == {"value3": "3"}
    assert config.section3.value3() == "3"


def test_file_does_not_exist(config):
    config.from_ini("./does_not_exist.ini")
    assert config() == {}


@mark.parametrize("config_type", ["strict"])
def test_file_does_not_exist_strict_mode(config):
    with raises(IOError):
        config.from_ini("./does_not_exist.ini")


def test_option_file_does_not_exist(config):
    config.option.from_ini("does_not_exist.ini")
    assert config.option.undefined() is None


@mark.parametrize("config_type", ["strict"])
def test_option_file_does_not_exist_strict_mode(config):
    with raises(IOError):
        config.option.from_ini("./does_not_exist.ini")


def test_required_file_does_not_exist(config):
    with raises(IOError):
        config.from_ini("./does_not_exist.ini", required=True)


def test_required_option_file_does_not_exist(config):
    with raises(IOError):
        config.option.from_ini("./does_not_exist.ini", required=True)


@mark.parametrize("config_type", ["strict"])
def test_not_required_file_does_not_exist_strict_mode(config):
    config.from_ini("./does_not_exist.ini", required=False)
    assert config() == {}


@mark.parametrize("config_type", ["strict"])
def test_not_required_option_file_does_not_exist_strict_mode(config):
    config.option.from_ini("./does_not_exist.ini", required=False)
    with raises(errors.Error):
        config.option()



"""Tests for container cls with custom string classes as attribute names.

See: https://github.com/ets-labs/python-dependency-injector/issues/479
"""

from dependency_injector import containers, providers
from pytest import fixture, raises


class CustomString(str):
    pass


class CustomClass:
    thing = None


class Container(containers.DeclarativeContainer):
    pass


@fixture
def provider():
    return providers.Provider()


def test_setattr(provider):
    setattr(Container, CustomString("test_attr"), provider)
    assert Container.test_attr is provider


def test_delattr():
    setattr(Container, CustomString("test_attr"), provider)
    delattr(Container, CustomString("test_attr"))
    with raises(AttributeError):
        Container.test_attr
